# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ..datasource import tensor as astensor
from ..core import TENSOR_TYPE


def shuffle(random_state, x, axis=0):
    r"""
    Modify a sequence in-place by shuffling its contents.
    The order of sub-arrays is changed but their contents remains the same.

    Parameters
    ----------
    x : array_like
        The array or list to be shuffled.
    axis : int, optional
        The axis which `x` is shuffled along. Default is 0.

    Returns
    -------
    None

    Examples
    --------
    >>> import mars.tensor as mt
    >>> rng = mt.random.RandomState()
    >>> arr = mt.arange(10)
    >>> rng.shuffle(arr)
    >>> arr.execute()
    array([0, 1, 4, 2, 8, 6, 5, 9, 3, 7]) # random

    >>> arr = mt.arange(9).reshape((3, 3))
    >>> rng.shuffle(arr)
    >>> arr.execute()
    array([[6, 7, 8], # random
           [0, 1, 2],
           [3, 4, 5]])
    """
    from .permutation import permutation

    if isinstance(x, (list, np.ndarray, TENSOR_TYPE)):
        x = astensor(x)
    else:
        raise TypeError("x should be list, numpy ndarray or tensor")

    ret = permutation(random_state, x, axis=axis)
    x.data = ret.data



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from typing import Callable, Dict, List, Union

from ....utils import parse_readable_size
from ...web import web_api, MarsServiceWebAPIHandler, MarsWebAPIClientMixin
from ..core import SessionInfo
from .core import AbstractSessionAPI


def _encode_size(size: Union[str, Dict[str, List[int]]]) -> str:
    if not isinstance(size, dict):
        return size
    else:
        return ",".join(f"{k}={v}" for k, v in size.items())


def _decode_size(encoded: str) -> Union[int, str, Dict[str, Union[int, List[int]]]]:
    if not encoded:
        return 0
    if "," not in encoded and "=" not in encoded:
        try:
            return int(encoded)
        except ValueError:
            return int(parse_readable_size(encoded)[0])
    else:
        ret = dict()
        for kv in encoded.split(","):
            k, v = kv.split("=", 1)
            ret[k] = int(parse_readable_size(v)[0])
        return ret


class SessionWebAPIBaseHandler(MarsServiceWebAPIHandler):
    async def _get_oscar_session_api(self):
        from .oscar import SessionAPI

        return await self._get_api_by_key(SessionAPI, "Session", with_key_arg=False)


class SessionWebAPIHandler(SessionWebAPIBaseHandler):
    @classmethod
    def get_root_pattern(cls):
        return "/api/session(?:/(?P<sub_path>[^/]*)$|$)"

    @web_api("(?P<session_id>[^/]+)", method="put")
    async def create_session(self, session_id: str):
        oscar_api = await self._get_oscar_session_api()
        addr = await oscar_api.create_session(session_id)
        self.write(addr)

    @web_api("(?P<session_id>[^/]+)", method="delete")
    async def delete_session(self, session_id: str):
        oscar_api = await self._get_oscar_session_api()
        await oscar_api.delete_session(session_id)

    @web_api("", method="delete")
    async def delete_all_sessions(self):
        oscar_api = await self._get_oscar_session_api()
        await oscar_api.delete_all_sessions()

    @web_api(
        "(?P<session_id>[^/]+)", method="get", arg_filter={"action": "check_exist"}
    )
    async def has_session(self, session_id: str):
        oscar_api = await self._get_oscar_session_api()
        res = await oscar_api.has_session(session_id)
        self.write("1" if res else "0")

    @web_api(
        "(?P<session_id>[^/]*)",
        method="get",
        arg_filter={"action": "get_last_idle_time"},
    )
    async def get_last_idle_time(self, session_id: str):
        session_id = session_id or None
        oscar_api = await self._get_oscar_session_api()
        res = await oscar_api.get_last_idle_time(session_id)
        self.write(str(res) if res else "")

    @web_api("", method="get")
    async def get_sessions(self):
        oscar_api = await self._get_oscar_session_api()
        res = await oscar_api.get_sessions()
        self.write(
            json.dumps({"sessions": [{"session_id": info.session_id} for info in res]})
        )


class SessionWebLogAPIHandler(SessionWebAPIBaseHandler):
    _root_pattern = "/api/session/(?P<session_id>[^/]+)/op/(?P<op_key>[^/]+)/log"

    @web_api("", method="get")
    async def fetch_tileable_op_logs(self, session_id: str, op_key: str):
        oscar_api = await self._get_oscar_session_api()
        offsets = _decode_size(self.get_argument("offsets", None))
        sizes = _decode_size(self.get_argument("sizes", None))
        log_result = await oscar_api.fetch_tileable_op_logs(
            session_id, op_key, offsets, sizes
        )
        self.write(json.dumps(log_result))


web_handlers = {
    SessionWebAPIHandler.get_root_pattern(): SessionWebAPIHandler,
    SessionWebLogAPIHandler.get_root_pattern(): SessionWebLogAPIHandler,
}


class WebSessionAPI(AbstractSessionAPI, MarsWebAPIClientMixin):
    def __init__(self, address: str, request_rewriter: Callable = None):
        self._address = address.rstrip("/")
        self.request_rewriter = request_rewriter

    async def get_sessions(self) -> List[SessionInfo]:
        addr = f"{self._address}/api/session"
        res = await self._request_url("GET", addr)
        res_obj = json.loads(res.body.decode())
        return [SessionInfo(**kw) for kw in res_obj["sessions"]]

    async def create_session(self, session_id: str) -> str:
        addr = f"{self._address}/api/session/{session_id}"
        res = await self._request_url(path=addr, method="PUT", data=b"")
        return res.body.decode()

    async def delete_session(self, session_id: str):
        addr = f"{self._address}/api/session/{session_id}"
        await self._request_url(path=addr, method="DELETE")

    async def delete_all_sessions(self):
        addr = f"{self._address}/api/session"
        await self._request_url(path=addr, method="DELETE")

    async def has_session(self, session_id: str):
        addr = f"{self._address}/api/session/{session_id}"
        params = dict(action="check_exist")
        res = await self._request_url("GET", addr, params=params)
        return bool(int(res.body.decode()))

    async def get_last_idle_time(
        self, session_id: Union[str, None] = None
    ) -> Union[float, None]:
        session_id = session_id or ""
        addr = f"{self._address}/api/session/{session_id}"
        params = dict(action="get_last_idle_time")
        res = await self._request_url("GET", addr, params=params)
        content = res.body.decode()
        return float(content) if content else None

    async def fetch_tileable_op_logs(
        self,
        session_id: str,
        tileable_op_key: str,
        chunk_op_key_to_offsets: Dict[str, List[int]],
        chunk_op_key_to_sizes: Dict[str, List[int]],
    ) -> Dict:
        addr = f"{self._address}/api/session/{session_id}/op/{tileable_op_key}/log"
        params = dict(
            offsets=_encode_size(chunk_op_key_to_offsets),
            sizes=_encode_size(chunk_op_key_to_sizes),
        )
        res = await self._request_url("GET", addr, params=params)
        return json.loads(res.body.decode())



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import opcodes as OperandDef
from ...serialization.serializables import (
    FieldTypes,
    DictField,
    TupleField,
    StringField,
    Int64Field,
)
from ...lib.sparse.core import sps
from ...lib.sparse import SparseNDArray
from ..core import TensorOrder
from .core import TensorNoInput


class TensorTileDBDataSource(TensorNoInput):
    _op_type_ = OperandDef.TENSOR_FROM_TILEDB

    _tiledb_config = DictField("tiledb_config")
    # URI of array to open
    _tiledb_uri = StringField("tiledb_uri")
    # tiledb dim start
    _tiledb_dim_starts = TupleField("tiledb_dim_starts", FieldTypes.int64)
    # encryption key to decrypt if provided
    _tiledb_key = StringField("tiledb_key")
    # open array at a given timestamp if provided
    _tiledb_timestamp = Int64Field("tiledb_timestamp")
    _axis_offsets = TupleField("axis_offsets", FieldTypes.int64)

    def __init__(
        self,
        tiledb_config=None,
        tiledb_uri=None,
        tiledb_dim_starts=None,
        tiledb_key=None,
        tiledb_timstamp=None,
        **kw
    ):
        super().__init__(
            _tiledb_config=tiledb_config,
            _tiledb_uri=tiledb_uri,
            _tiledb_dim_starts=tiledb_dim_starts,
            _tiledb_key=tiledb_key,
            _tiledb_timestamp=tiledb_timstamp,
            **kw
        )

    @property
    def tiledb_config(self):
        return self._tiledb_config

    @property
    def tiledb_uri(self):
        return self._tiledb_uri

    @property
    def tiledb_dim_starts(self):
        return self._tiledb_dim_starts

    @property
    def tiledb_key(self):
        return self._tiledb_key

    @property
    def tiledb_timestamp(self):
        return self._tiledb_timestamp

    @property
    def axis_offsets(self):
        return self._axis_offsets

    def to_chunk_op(self, *args):
        _, chunk_idx, nsplits = args
        chunk_op = super().to_chunk_op(*args)
        axis_offsets = []
        for axis, idx in enumerate(chunk_idx):
            axis_offsets.append(sum(nsplits[axis][:idx]))
        chunk_op._axis_offsets = tuple(axis_offsets)
        return chunk_op

    @classmethod
    def execute(cls, ctx, op):
        import tiledb

        chunk = op.outputs[0]
        from ..array_utils import array_module
        from ..utils import get_tiledb_ctx

        xp = array_module(op.gpu)

        axis_offsets = [
            offset + dim_start
            for offset, dim_start in zip(op.axis_offsets, op.tiledb_dim_starts)
        ]
        tiledb_ctx = get_tiledb_ctx(op.tiledb_config)
        uri = op.tiledb_uri
        key = op.tiledb_key
        timestamp = op.tiledb_timestamp

        slcs = []
        for axis in range(chunk.ndim):
            axis_offset = axis_offsets[axis]
            axis_length = chunk.shape[axis]
            slcs.append(slice(axis_offset, axis_offset + axis_length))

        if not op.sparse:
            # read dense array from tiledb
            with tiledb.DenseArray(
                uri=uri, ctx=tiledb_ctx, key=key, timestamp=timestamp
            ) as tiledb_arr:
                ctx[chunk.key] = tiledb_arr[tuple(slcs)]
        else:
            # read sparse array from tiledb
            with tiledb.SparseArray(
                uri=uri, ctx=tiledb_ctx, key=key, timestamp=timestamp
            ) as tiledb_arr:
                if tiledb_arr.ndim > 2:
                    raise NotImplementedError(
                        "Does not support to read array with more than 2 dimensions"
                    )

                data = tiledb_arr[tuple(slcs)]
                coords = data["coords"]

                value = data[tiledb_arr.attr(0).name]
                if tiledb_arr.ndim == 2:
                    # 2-d
                    ij = tuple(
                        coords[tiledb_arr.domain.dim(k).name] - axis_offsets[k]
                        for k in range(tiledb_arr.ndim)
                    )
                    spmatrix = sps.coo_matrix((value, ij), shape=chunk.shape)
                    ctx[chunk.key] = SparseNDArray(spmatrix)
                else:
                    # 1-d
                    ij = (
                        xp.zeros(coords.shape),
                        coords[tiledb_arr.domain.dim(0).name] - axis_offsets[0],
                    )
                    spmatrix = sps.coo_matrix((value, ij), shape=(1,) + chunk.shape)
                    ctx[chunk.key] = SparseNDArray(spmatrix, shape=chunk.shape)


def fromtiledb(uri, ctx=None, key=None, timestamp=None, gpu=None):
    import tiledb

    raw_ctx = ctx
    if raw_ctx is None:
        ctx = tiledb.Ctx()

    # get metadata from tiledb
    try:
        tiledb_arr = tiledb.DenseArray(uri=uri, ctx=ctx, key=key, timestamp=timestamp)
        sparse = False
    except ValueError:
        # if the array is not dense, ValueError will be raised by tiledb
        tiledb_arr = tiledb.SparseArray(uri=uri, ctx=ctx, key=key, timestamp=timestamp)
        sparse = True

    if tiledb_arr.nattr > 1:
        raise NotImplementedError(
            "Does not supported TileDB array schema with more than 1 attr"
        )
    tiledb_dim_starts = tuple(
        tiledb_arr.domain.dim(j).domain[0].item() for j in range(tiledb_arr.ndim)
    )
    if any(isinstance(s, float) for s in tiledb_dim_starts):
        raise ValueError(
            "Does not support TileDB array schema whose dimensions has float domain"
        )

    dtype = tiledb_arr.attr(0).dtype
    tiledb_config = None if raw_ctx is None else ctx.config().dict()
    tensor_order = (
        TensorOrder.C_ORDER
        if tiledb_arr.schema.cell_order == "row-major"
        else TensorOrder.F_ORDER
    )
    op = TensorTileDBDataSource(
        tiledb_config=tiledb_config,
        tiledb_uri=uri,
        tiledb_key=key,
        tiledb_timstamp=timestamp,
        tiledb_dim_starts=tiledb_dim_starts,
        gpu=gpu,
        sparse=sparse,
        dtype=dtype,
    )
    chunk_size = tuple(
        int(tiledb_arr.domain.dim(i).tile) for i in range(tiledb_arr.domain.ndim)
    )
    return op(tiledb_arr.shape, chunk_size=chunk_size, order=tensor_order)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from functools import partial

import numpy as np

from ... import execute as _execute, fetch as _fetch
from ... import tensor as mt
from ...utils import cache_tileables
from ..preprocessing import label_binarize
from ..utils._encode import _encode, _unique
from ..utils.checks import assert_all_finite
from ..utils.core import sort_by
from ..utils.multiclass import type_of_target
from ..utils.validation import check_array, check_consistent_length, column_or_1d
from ._base import _average_binary_score, _average_multiclass_ovo_score


def auc(x, y, session=None, run_kwargs=None):
    """Compute Area Under the Curve (AUC) using the trapezoidal rule

    This is a general function, given points on a curve.  For computing the
    area under the ROC-curve, see :func:`roc_auc_score`.  For an alternative
    way to summarize a precision-recall curve, see
    :func:`average_precision_score`.

    Parameters
    ----------
    x : tensor, shape = [n]
        x coordinates. These must be either monotonic increasing or monotonic
        decreasing.
    y : tensor, shape = [n]
        y coordinates.

    Returns
    -------
    auc : tensor, with float value

    Examples
    --------
    >>> import mars.tensor as mt
    >>> from mars.learn import metrics
    >>> y = mt.array([1, 1, 2, 2])
    >>> pred = mt.array([0.1, 0.4, 0.35, 0.8])
    >>> fpr, tpr, thresholds = metrics.roc_curve(y, pred, pos_label=2)
    >>> metrics.auc(fpr, tpr)
    0.75

    See also
    --------
    roc_auc_score : Compute the area under the ROC curve
    average_precision_score : Compute average precision from prediction scores
    precision_recall_curve :
        Compute precision-recall pairs for different probability thresholds
    """
    check_consistent_length(x, y)
    x = column_or_1d(x)
    y = column_or_1d(y)

    if x.shape[0] < 2:
        raise ValueError(
            "At least 2 points are needed to compute"
            f" area under curve, but x.shape = {x.shape}"
        )

    direction = 1
    dx = mt.diff(x)
    any_dx_lt_0 = mt.any(dx < 0)
    all_dx_le_0 = mt.all(dx <= 0)
    mt.ExecutableTuple([x, any_dx_lt_0, all_dx_le_0]).execute(
        session=session, **(run_kwargs or dict())
    )
    if any_dx_lt_0.fetch(session=session):
        if all_dx_le_0.fetch(session=session):
            direction = -1
        else:
            x_data = x.fetch(session=session)
            raise ValueError(f"x is neither increasing nor decreasing : {x_data}.")

    area = direction * mt.trapz(y, x)
    return area.execute(session=session, **(run_kwargs or dict()))


def _binary_clf_curve(
    y_true, y_score, pos_label=None, sample_weight=None, session=None, run_kwargs=None
):
    """Calculate true and false positives per binary classification threshold.

    Parameters
    ----------
    y_true : tensor, shape = [n_samples]
        True targets of binary classification

    y_score : tensor, shape = [n_samples]
        Estimated probabilities or decision function

    pos_label : int or str, default=None
        The label of the positive class

    sample_weight : array-like of shape (n_samples,), default=None
        Sample weights.

    Returns
    -------
    fps : tensor, shape = [n_thresholds]
        A count of false positives, at index i being the number of negative
        samples assigned a score >= thresholds[i]. The total number of
        negative samples is equal to fps[-1] (thus true negatives are given by
        fps[-1] - fps).

    tps : tensor, shape = [n_thresholds <= len(mt.unique(y_score))]
        An increasing count of true positives, at index i being the number
        of positive samples assigned a score >= thresholds[i]. The total
        number of positive samples is equal to tps[-1] (thus false negatives
        are given by tps[-1] - tps).

    thresholds : tensor, shape = [n_thresholds]
        Decreasing score values.
    """
    y_type = type_of_target(y_true).to_numpy(session=session, **(run_kwargs or dict()))
    if not (y_type == "binary" or (y_type == "multiclass" and pos_label is not None)):
        raise ValueError(f"{y_type} format is not supported")

    check_consistent_length(
        y_true, y_score, sample_weight, session=session, **(run_kwargs or dict())
    )
    y_true = column_or_1d(y_true)
    y_score = column_or_1d(y_score)
    y_true = assert_all_finite(y_true, check_only=False)
    y_score = assert_all_finite(y_score, check_only=False)

    cache_tileables(y_true, y_score)

    if sample_weight is not None:
        sample_weight = column_or_1d(sample_weight)

    # ensure binary classification if pos_label is not specified
    # classes.dtype.kind in ('O', 'U', 'S') is required to avoid
    # triggering a FutureWarning by calling np.array_equal(a, b)
    # when elements in the two arrays are not comparable.
    classes = mt.unique(y_true, aggregate_size=1).to_numpy(
        session=session, **(run_kwargs or dict())
    )
    if pos_label is None and (
        classes.dtype.kind in ("O", "U", "S")
        or not (
            np.array_equal(classes, [0, 1])
            or np.array_equal(classes, [-1, 1])
            or np.array_equal(classes, [0])
            or np.array_equal(classes, [-1])
            or np.array_equal(classes, [1])
        )
    ):
        classes_repr = ", ".join(repr(c) for c in classes)
        raise ValueError(
            f"y_true takes value in {{{classes_repr}}} and "
            "pos_label is not specified: either make y_true "
            "take value in {{0, 1}} or {{-1, 1}} or "
            "pass pos_label explicitly."
        )
    elif pos_label is None:
        pos_label = 1.0

    # make y_true a boolean vector
    y_true = y_true == pos_label

    # sort scores and corresponding truth values
    # original implementation adopted from sklearn:
    # """
    # desc_score_indices = mt.argsort(y_score, kind="mergesort")[::-1]
    # y_score = y_score[desc_score_indices]
    # y_true = y_true[desc_score_indices]
    # if sample_weight is not None:
    #     weight = sample_weight[desc_score_indices]
    # else:
    #     weight = 1.0
    # """
    # since fancy indexing is a heavy operation, we try to use DataFrame to sort
    to_sort = [y_score, y_true]
    if sample_weight is not None:
        to_sort.append(sample_weight)
    to_sort = sort_by(to_sort, y_score, ascending=False)
    y_score, y_true = to_sort[:2]
    if sample_weight is not None:
        weight = to_sort[-1]
    else:
        weight = 1.0

    # y_score typically has many tied values. Here we extract
    # the indices associated with the distinct values. We also
    # concatenate a value for the end of the curve.
    distinct_value_indices = mt.where(mt.diff(y_score))[0]
    threshold_idxs = mt.r_[distinct_value_indices, y_true.size - 1]

    # accumulate the true positives with decreasing threshold
    # raw tps from sklearn implementation
    # we try to perform only one fancy index
    # tps = (y_true * weight).cumsum()[threshold_idxs]
    temp_tps = (y_true * weight).cumsum()
    if sample_weight is not None:
        # express fps as a cumsum to ensure fps is increasing even in
        # the presence of floating point errors
        # fps = ((1 - y_true) * weight).cumsum()[threshold_idxs]
        temp_fps = ((1 - y_true) * weight).cumsum()
        tps, fps, thresholds = mt.stack([temp_tps, temp_fps, y_score])[
            :, threshold_idxs
        ]

    else:
        tps, thresholds = mt.stack([temp_tps, y_score])[:, threshold_idxs]
        fps = 1 + threshold_idxs - tps
    return _execute([fps, tps, thresholds], session=session, **(run_kwargs or dict()))


def _binary_roc_auc_score(
    y_true, y_score, sample_weight=None, max_fpr=None, session=None, run_kwargs=None
):
    """Binary roc auc score."""

    from numpy import interp

    if len(mt.unique(y_true).execute()) != 2:
        raise ValueError(
            "Only one class present in y_true. ROC AUC score "
            "is not defined in that case."
        )

    fpr, tpr, _ = roc_curve(
        y_true,
        y_score,
        sample_weight=sample_weight,
        session=session,
        run_kwargs=run_kwargs,
    )
    fpr, tpr = mt.ExecutableTuple([fpr, tpr]).fetch(session=session)

    if max_fpr is None or max_fpr == 1:
        return auc(fpr, tpr, session=session, run_kwargs=run_kwargs).fetch(
            session=session
        )
    if max_fpr <= 0 or max_fpr > 1:
        raise ValueError(f"Expected max_fpr in range (0, 1], got: {max_fpr}")

    # Add a single point at max_fpr by linear interpolation
    stop = (
        mt.searchsorted(fpr, max_fpr, "right")
        .execute(session=session, **(run_kwargs or dict()))
        .fetch(session=session)
    )
    x_interp = [fpr[stop - 1], fpr[stop]]
    y_interp = [tpr[stop - 1], tpr[stop]]
    tpr = list(tpr[:stop])
    tpr.append(interp(max_fpr, x_interp, y_interp))
    fpr = list(fpr[:stop])
    fpr.append(max_fpr)
    partial_auc = auc(fpr, tpr, session=session, run_kwargs=run_kwargs)

    # McClish correction: standardize result to be 0.5 if non-discriminant
    # and 1 if maximal
    min_area = 0.5 * max_fpr**2
    max_area = max_fpr
    return 0.5 * (
        1 + (partial_auc.fetch(session=session) - min_area) / (max_area - min_area)
    )


def roc_auc_score(
    y_true,
    y_score,
    *,
    average="macro",
    sample_weight=None,
    max_fpr=None,
    multi_class="raise",
    labels=None,
    session=None,
    run_kwargs=None,
):
    """
    Compute Area Under the Receiver Operating Characteristic Curve (ROC AUC)
    from prediction scores.

    Note: this implementation can be used with binary, multiclass and
    multilabel classification, but some restrictions apply (see Parameters).

    Read more in the :ref:`User Guide <roc_metrics>`.

    Parameters
    ----------
    y_true : array-like of shape (n_samples,) or (n_samples, n_classes)
        True labels or binary label indicators. The binary and multiclass cases
        expect labels with shape (n_samples,) while the multilabel case expects
        binary label indicators with shape (n_samples, n_classes).

    y_score : array-like of shape (n_samples,) or (n_samples, n_classes)
        Target scores.

        * In the binary case, it corresponds to an array of shape
          `(n_samples,)`. Both probability estimates and non-thresholded
          decision values can be provided. The probability estimates correspond
          to the **probability of the class with the greater label**,
          i.e. `estimator.classes_[1]` and thus
          `estimator.predict_proba(X, y)[:, 1]`. The decision values
          corresponds to the output of `estimator.decision_function(X, y)`.
          See more information in the :ref:`User guide <roc_auc_binary>`;
        * In the multiclass case, it corresponds to an array of shape
          `(n_samples, n_classes)` of probability estimates provided by the
          `predict_proba` method. The probability estimates **must**
          sum to 1 across the possible classes. In addition, the order of the
          class scores must correspond to the order of ``labels``,
          if provided, or else to the numerical or lexicographical order of
          the labels in ``y_true``. See more information in the
          :ref:`User guide <roc_auc_multiclass>`;
        * In the multilabel case, it corresponds to an array of shape
          `(n_samples, n_classes)`. Probability estimates are provided by the
          `predict_proba` method and the non-thresholded decision values by
          the `decision_function` method. The probability estimates correspond
          to the **probability of the class with the greater label for each
          output** of the classifier. See more information in the
          :ref:`User guide <roc_auc_multilabel>`.

    average : {'micro', 'macro', 'samples', 'weighted'} or None, \
            default='macro'
        If ``None``, the scores for each class are returned. Otherwise,
        this determines the type of averaging performed on the data:
        Note: multiclass ROC AUC currently only handles the 'macro' and
        'weighted' averages.

        ``'micro'``:
            Calculate metrics globally by considering each element of the label
            indicator matrix as a label.
        ``'macro'``:
            Calculate metrics for each label, and find their unweighted
            mean.  This does not take label imbalance into account.
        ``'weighted'``:
            Calculate metrics for each label, and find their average, weighted
            by support (the number of true instances for each label).
        ``'samples'``:
            Calculate metrics for each instance, and find their average.

        Will be ignored when ``y_true`` is binary.

    sample_weight : array-like of shape (n_samples,), default=None
        Sample weights.

    max_fpr : float > 0 and <= 1, default=None
        If not ``None``, the standardized partial AUC [2]_ over the range
        [0, max_fpr] is returned. For the multiclass case, ``max_fpr``,
        should be either equal to ``None`` or ``1.0`` as AUC ROC partial
        computation currently is not supported for multiclass.

    multi_class : {'raise', 'ovr', 'ovo'}, default='raise'
        Only used for multiclass targets. Determines the type of configuration
        to use. The default value raises an error, so either
        ``'ovr'`` or ``'ovo'`` must be passed explicitly.

        ``'ovr'``:
            Stands for One-vs-rest. Computes the AUC of each class
            against the rest [3]_ [4]_. This
            treats the multiclass case in the same way as the multilabel case.
            Sensitive to class imbalance even when ``average == 'macro'``,
            because class imbalance affects the composition of each of the
            'rest' groupings.
        ``'ovo'``:
            Stands for One-vs-one. Computes the average AUC of all
            possible pairwise combinations of classes [5]_.
            Insensitive to class imbalance when
            ``average == 'macro'``.

    labels : array-like of shape (n_classes,), default=None
        Only used for multiclass targets. List of labels that index the
        classes in ``y_score``. If ``None``, the numerical or lexicographical
        order of the labels in ``y_true`` is used.

    Returns
    -------
    auc : float

    References
    ----------
    .. [1] `Wikipedia entry for the Receiver operating characteristic
            <https://en.wikipedia.org/wiki/Receiver_operating_characteristic>`_

    .. [2] `Analyzing a portion of the ROC curve. McClish, 1989
            <https://www.ncbi.nlm.nih.gov/pubmed/2668680>`_

    .. [3] Provost, F., Domingos, P. (2000). Well-trained PETs: Improving
           probability estimation trees (Section 6.2), CeDER Working Paper
           #IS-00-04, Stern School of Business, New York University.

    .. [4] `Fawcett, T. (2006). An introduction to ROC analysis. Pattern
            Recognition Letters, 27(8), 861-874.
            <https://www.sciencedirect.com/science/article/pii/S016786550500303X>`_

    .. [5] `Hand, D.J., Till, R.J. (2001). A Simple Generalisation of the Area
            Under the ROC Curve for Multiple Class Classification Problems.
            Machine Learning, 45(2), 171-186.
            <http://link.springer.com/article/10.1023/A:1010920819831>`_

    See Also
    --------
    average_precision_score : Area under the precision-recall curve.
    roc_curve : Compute Receiver operating characteristic (ROC) curve.
    RocCurveDisplay.from_estimator : Plot Receiver Operating Characteristic
        (ROC) curve given an estimator and some data.
    RocCurveDisplay.from_predictions : Plot Receiver Operating Characteristic
        (ROC) curve given the true and predicted values.

    Examples
    --------
    Binary case:

    >>> from sklearn.datasets import load_breast_cancer
    >>> from sklearn.linear_model import LogisticRegression
    >>> from mars.learn.metrics import roc_auc_score
    >>> X, y = load_breast_cancer(return_X_y=True)
    >>> clf = LogisticRegression(solver="liblinear", random_state=0).fit(X, y)
    >>> roc_auc_score(y, clf.predict_proba(X)[:, 1])
    0.99...
    >>> roc_auc_score(y, clf.decision_function(X))
    0.99...

    Multiclass case:

    >>> from sklearn.datasets import load_iris
    >>> X, y = load_iris(return_X_y=True)
    >>> clf = LogisticRegression(solver="liblinear").fit(X, y)
    >>> roc_auc_score(y, clf.predict_proba(X), multi_class='ovr')
    0.99...

    Multilabel case:

    >>> import numpy as np
    >>> from sklearn.datasets import make_multilabel_classification
    >>> from sklearn.multioutput import MultiOutputClassifier
    >>> X, y = make_multilabel_classification(random_state=0)
    >>> clf = MultiOutputClassifier(clf).fit(X, y)
    >>> # get a list of n_output containing probability arrays of shape
    >>> # (n_samples, n_classes)
    >>> y_pred = clf.predict_proba(X)
    >>> # extract the positive columns for each output
    >>> y_pred = np.transpose([pred[:, 1] for pred in y_pred])
    >>> roc_auc_score(y, y_pred, average=None)
    array([0.82..., 0.86..., 0.94..., 0.85... , 0.94...])
    >>> from sklearn.linear_model import RidgeClassifierCV
    >>> clf = RidgeClassifierCV().fit(X, y)
    >>> roc_auc_score(y, clf.decision_function(X), average=None)
    array([0.81..., 0.84... , 0.93..., 0.87..., 0.94...])
    """

    cache_tileables(y_true, y_score)

    y_type = type_of_target(y_true)
    y_true = check_array(y_true, ensure_2d=False, dtype=None)
    y_score = check_array(y_score, ensure_2d=False)
    _execute([y_type, y_true, y_score], session=session, **(run_kwargs or dict()))
    y_type = y_type.fetch(session=session)

    def execute(*args):
        result = [None] * len(args)
        to_execute = dict()
        for i, arg in enumerate(args):
            if hasattr(arg, "op"):
                to_execute[i] = arg
            else:
                result[i] = arg
        if to_execute:
            _execute(*to_execute.values(), session=session, **(run_kwargs or dict()))
        for i, e in to_execute.items():
            if e.isscalar():
                e = e.fetch(session=session)
            result[i] = e
        return result[0] if len(result) == 1 else result

    if y_type == "multiclass" or (
        y_type == "binary" and y_score.ndim == 2 and y_score.shape[1] > 2
    ):
        # do not support partial ROC computation for multiclass
        if max_fpr is not None and max_fpr != 1.0:
            raise ValueError(
                "Partial AUC computation not available in "
                "multiclass setting, 'max_fpr' must be"
                " set to `None`, received `max_fpr={0}` "
                "instead".format(max_fpr)
            )
        if multi_class == "raise":
            raise ValueError("multi_class must be in ('ovo', 'ovr')")
        return execute(
            _multiclass_roc_auc_score(
                y_true, y_score, labels, multi_class, average, sample_weight
            )
        )
    elif y_type == "binary":
        labels = mt.unique(y_true).execute(session=session, **(run_kwargs or dict()))
        y_true = label_binarize(y_true, classes=labels, execute=False)[:, 0]
        cache_tileables(y_true)
        return execute(
            _average_binary_score(
                partial(_binary_roc_auc_score, max_fpr=max_fpr),
                y_true,
                y_score,
                average,
                sample_weight=sample_weight,
            )
        )
    else:  # multilabel-indicator
        return execute(
            _average_binary_score(
                partial(_binary_roc_auc_score, max_fpr=max_fpr),
                y_true,
                y_score,
                average,
                sample_weight=sample_weight,
            )
        )


def _multiclass_roc_auc_score(
    y_true,
    y_score,
    labels,
    multi_class,
    average,
    sample_weight,
    session=None,
    run_kwargs=None,
):
    # validation of the input y_score
    if not mt.allclose(1, y_score.sum(axis=1)).to_numpy(
        session=session, **(run_kwargs or dict())
    ):  # pragma: no cover
        raise ValueError(
            "Target scores need to be probabilities for multiclass "
            "roc_auc, i.e. they should sum up to 1.0 over classes"
        )

    # validation for multiclass parameter specifications
    average_options = ("macro", "weighted")
    if average not in average_options:
        raise ValueError(
            "average must be one of {0} for multiclass problems".format(average_options)
        )

    multiclass_options = ("ovo", "ovr")
    if multi_class not in multiclass_options:
        raise ValueError(
            "multi_class='{0}' is not supported "
            "for multiclass ROC AUC, multi_class must be "
            "in {1}".format(multi_class, multiclass_options)
        )

    if labels is not None:
        labels = column_or_1d(labels).to_numpy(
            session=session, **(run_kwargs or dict())
        )
        classes = _unique(labels).to_numpy(session=session, **(run_kwargs or dict()))
        if len(classes) != len(labels):
            raise ValueError("Parameter 'labels' must be unique")
        if not np.array_equal(classes, labels):
            raise ValueError("Parameter 'labels' must be ordered")
        if len(classes) != y_score.shape[1]:
            raise ValueError(
                "Number of given labels, {0}, not equal to the number "
                "of columns in 'y_score', {1}".format(len(classes), y_score.shape[1])
            )
        if len(
            mt.setdiff1d(y_true, classes).execute(
                session=session, **(run_kwargs or dict())
            )
        ):
            raise ValueError("'y_true' contains labels not in parameter 'labels'")
    else:
        classes = _unique(y_true).execute(session=session, **(run_kwargs or dict()))
        if len(classes) != y_score.shape[1]:
            raise ValueError(
                "Number of classes in y_true not equal to the number of "
                "columns in 'y_score'"
            )

    if multi_class == "ovo":
        if sample_weight is not None:
            raise ValueError(
                "sample_weight is not supported "
                "for multiclass one-vs-one ROC AUC, "
                "'sample_weight' must be None in this case."
            )
        y_true_encoded = _encode(y_true, uniques=classes)
        # Hand & Till (2001) implementation (ovo)
        return _average_multiclass_ovo_score(
            _binary_roc_auc_score,
            y_true_encoded,
            y_score,
            average=average,
            session=session,
            run_kwargs=run_kwargs,
        )
    else:
        # ovr is same as multi-label
        y_true_multilabel = label_binarize(y_true, classes=classes, execute=False)
        return _average_binary_score(
            _binary_roc_auc_score,
            y_true_multilabel,
            y_score,
            average,
            sample_weight=sample_weight,
            session=session,
            run_kwargs=run_kwargs,
        )


def roc_curve(
    y_true,
    y_score,
    pos_label=None,
    sample_weight=None,
    drop_intermediate=True,
    session=None,
    run_kwargs=None,
):
    """Compute Receiver operating characteristic (ROC)

    Note: this implementation is restricted to the binary classification task.

    Read more in the :ref:`User Guide <roc_metrics>`.

    Parameters
    ----------

    y_true : tensor, shape = [n_samples]
        True binary labels. If labels are not either {-1, 1} or {0, 1}, then
        pos_label should be explicitly given.

    y_score : tensor, shape = [n_samples]
        Target scores, can either be probability estimates of the positive
        class, confidence values, or non-thresholded measure of decisions
        (as returned by "decision_function" on some classifiers).

    pos_label : int or str, default=None
        The label of the positive class.
        When ``pos_label=None``, if y_true is in {-1, 1} or {0, 1},
        ``pos_label`` is set to 1, otherwise an error will be raised.

    sample_weight : array-like of shape (n_samples,), default=None
        Sample weights.

    drop_intermediate : boolean, optional (default=True)
        Whether to drop some suboptimal thresholds which would not appear
        on a plotted ROC curve. This is useful in order to create lighter
        ROC curves.

        .. versionadded:: 0.17
           parameter *drop_intermediate*.

    Returns
    -------
    fpr : tensor, shape = [>2]
        Increasing false positive rates such that element i is the false
        positive rate of predictions with score >= thresholds[i].

    tpr : tensor, shape = [>2]
        Increasing true positive rates such that element i is the true
        positive rate of predictions with score >= thresholds[i].

    thresholds : tensor, shape = [n_thresholds]
        Decreasing thresholds on the decision function used to compute
        fpr and tpr. `thresholds[0]` represents no instances being predicted
        and is arbitrarily set to `max(y_score) + 1`.

    See also
    --------
    roc_auc_score : Compute the area under the ROC curve

    Notes
    -----
    Since the thresholds are sorted from low to high values, they
    are reversed upon returning them to ensure they correspond to both ``fpr``
    and ``tpr``, which are sorted in reversed order during their calculation.

    References
    ----------
    .. [1] `Wikipedia entry for the Receiver operating characteristic
            <https://en.wikipedia.org/wiki/Receiver_operating_characteristic>`_

    .. [2] Fawcett T. An introduction to ROC analysis[J]. Pattern Recognition
           Letters, 2006, 27(8):861-874.

    Examples
    --------
    >>> import mars.tensor as mt
    >>> from mars.learn import metrics
    >>> y = mt.array([1, 1, 2, 2])
    >>> scores = mt.array([0.1, 0.4, 0.35, 0.8])
    >>> fpr, tpr, thresholds = metrics.roc_curve(y, scores, pos_label=2)
    >>> fpr
    array([0. , 0. , 0.5, 0.5, 1. ])
    >>> tpr
    array([0. , 0.5, 0.5, 1. , 1. ])
    >>> thresholds
    array([1.8 , 0.8 , 0.4 , 0.35, 0.1 ])

    """
    from sklearn.exceptions import UndefinedMetricWarning

    cache_tileables(y_true, y_score)

    fps, tps, thresholds = _binary_clf_curve(
        y_true,
        y_score,
        pos_label=pos_label,
        sample_weight=sample_weight,
        session=session,
        run_kwargs=run_kwargs,
    )

    # Attempt to drop thresholds corresponding to points in between and
    # collinear with other points. These are always suboptimal and do not
    # appear on a plotted ROC curve (and thus do not affect the AUC).
    # Here mt.diff(_, 2) is used as a "second derivative" to tell if there
    # is a corner at the point. Both fps and tps must be tested to handle
    # thresholds with multiple data points (which are combined in
    # _binary_clf_curve). This keeps all cases where the point should be kept,
    # but does not drop more complicated cases like fps = [1, 3, 7],
    # tps = [1, 2, 4]; there is no harm in keeping too many thresholds.
    if drop_intermediate and len(fps) > 2:
        optimal_idxs = mt.where(
            mt.r_[True, mt.logical_or(mt.diff(fps, 2), mt.diff(tps, 2)), True]
        )[0]
        # original implementation of sklearn:
        # """
        # fps = fps[optimal_idxs]
        # tps = tps[optimal_idxs]
        # thresholds = thresholds[optimal_idxs]
        # """
        # however, it's really a heavy operation to perform fancy index,
        # thus we put them together
        stacked = mt.stack([fps, tps, thresholds])
        fps, tps, thresholds = stacked[:, optimal_idxs]

    # Add an extra threshold position
    # to make sure that the curve starts at (0, 0)
    tps = mt.r_[0, tps]
    fps = mt.r_[0, fps]
    thresholds = mt.r_[thresholds[0] + 1, thresholds]

    last_fps = fps[-1]
    last_tps = tps[-1]
    _execute(
        [tps, fps, last_fps, last_tps, thresholds],
        session=session,
        **(run_kwargs or dict()),
    )
    last_fps, last_tps = _fetch([last_fps, last_tps], session=session)

    if last_fps <= 0:
        warnings.warn(
            "No negative samples in y_true, "
            "false positive value should be meaningless",
            UndefinedMetricWarning,
        )
        fpr = mt.repeat(mt.nan, fps.shape)
    else:
        fpr = fps / last_fps

    if last_tps <= 0:
        warnings.warn(
            "No positive samples in y_true, "
            "true positive value should be meaningless",
            UndefinedMetricWarning,
        )
        tpr = mt.repeat(mt.nan, tps.shape)
    else:
        tpr = tps / last_tps

    ret = mt.ExecutableTuple([fpr, tpr, thresholds]).execute(
        session=session, **(run_kwargs or dict())
    )
    return ret



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
from collections import defaultdict
from typing import Any, Dict, List, Union

from ... import oscar as mo
from ...storage import StorageLevel, get_storage_backend
from ...storage.core import StorageFileObject
from ...typing import BandType
from ...utils import calc_data_size, lazy_import
from ..cluster import ClusterAPI, StorageInfo
from ..meta import MetaAPI
from .core import (
    StorageQuotaActor,
    DataManagerActor,
    DataInfo,
    build_data_info,
    WrappedStorageFileObject,
)
from .errors import DataNotExist, NoDataToSpill

cupy = lazy_import("cupy")
cudf = lazy_import("cudf")

logger = logging.getLogger(__name__)


class StorageHandlerActor(mo.Actor):
    """
    Storage handler actor, provide methods like `get`, `put`, etc.
    This actor is stateful and created on worker's sub pools.
    """

    def __init__(
        self,
        storage_init_params: Dict,
        data_manager_ref: mo.ActorRefType[DataManagerActor],
        spill_manager_refs,
        quota_refs: Dict[StorageLevel, mo.ActorRefType[StorageQuotaActor]],
        band_name: str = "numa-0",
    ):
        from .spill import SpillManagerActor

        self._storage_init_params = storage_init_params
        self._data_manager_ref = data_manager_ref
        self._spill_manager_refs: Dict[
            StorageLevel, mo.ActorRefType[SpillManagerActor]
        ] = spill_manager_refs
        self._quota_refs = quota_refs
        self._band_name = band_name
        self._supervisor_address = None

    @classmethod
    def gen_uid(cls, band_name: str):
        return f"storage_handler_{band_name}"

    @property
    def highest_level(self):
        return min(self._quota_refs)

    async def __post_create__(self):
        self._clients = clients = dict()
        for backend, init_params in self._storage_init_params.items():
            logger.debug("Start storage %s with params %s", backend, init_params)
            storage_cls = get_storage_backend(backend)
            client = storage_cls(**init_params)
            for level in StorageLevel.__members__.values():
                if client.level & level:
                    clients[level] = client

    async def _get_data(self, data_info, conditions):
        if conditions is None:
            res = yield self._clients[data_info.level].get(data_info.object_id)
        else:
            try:
                res = yield self._clients[data_info.level].get(
                    data_info.object_id, conditions=conditions
                )
            except NotImplementedError:
                data = yield self._clients[data_info.level].get(data_info.object_id)
                try:
                    sliced_value = data.iloc[tuple(conditions)]
                except AttributeError:
                    sliced_value = data[tuple(conditions)]
                res = sliced_value
        raise mo.Return(res)

    @mo.extensible
    async def get(
        self,
        session_id: str,
        data_key: str,
        conditions: List = None,
        error: str = "raise",
    ):
        try:
            data_info = await self._data_manager_ref.get_data_info(
                session_id, data_key, self._band_name
            )
            data = yield self._get_data(data_info, conditions)
            raise mo.Return(data)
        except DataNotExist:
            if error == "raise":
                raise

    def _get_data_info(
        self,
        session_id: str,
        data_key: str,
        conditions: List = None,
        error: str = "raise",
    ):
        info = self._data_manager_ref.get_data_info.delay(
            session_id, data_key, self._band_name, error
        )
        return info, conditions

    @get.batch
    async def batch_get(self, args_list, kwargs_list):
        infos = []
        conditions_list = []
        for args, kwargs in zip(args_list, kwargs_list):
            info, conditions = self._get_data_info(*args, **kwargs)
            infos.append(info)
            conditions_list.append(conditions)
        data_infos = await self._data_manager_ref.get_data_info.batch(*infos)
        results = []
        for data_info, conditions in zip(data_infos, conditions_list):
            if data_info is None:
                results.append(None)
            else:
                result = yield self._get_data(data_info, conditions)
                results.append(result)
        raise mo.Return(results)

    def _get_default_level(self, obj):
        obj = obj[0] if isinstance(obj, (list, tuple)) else obj
        if self.highest_level != StorageLevel.GPU:
            return self.highest_level
        else:  # pragma: no cover
            if cudf is not None and isinstance(
                obj, (cudf.DataFrame, cudf.Series, cudf.Index)
            ):
                return StorageLevel.GPU
            elif cupy is not None and isinstance(obj, cupy.ndarray):
                return StorageLevel.GPU
            else:
                return StorageLevel.MEMORY

    @mo.extensible
    async def put(
        self, session_id: str, data_key: str, obj: object, level: StorageLevel = None
    ) -> DataInfo:
        if level is None:
            level = self._get_default_level(obj)
        size = await asyncio.to_thread(calc_data_size, obj)
        await self.request_quota_with_spill(level, size)
        object_info = await self._clients[level].put(obj)
        data_info = build_data_info(object_info, level, size, self._band_name)
        await self._data_manager_ref.put_data_info(
            session_id, data_key, data_info, object_info
        )
        if object_info.size is not None and data_info.memory_size != object_info.size:
            await self._quota_refs[level].update_quota(
                object_info.size - data_info.memory_size
            )
        await self.notify_spillable_space(level)
        return data_info

    @put.batch
    async def batch_put(self, args_list, kwargs_list):
        objs = []
        data_keys = []
        session_id = None
        level = last_level = None
        sizes = []
        for args, kwargs in zip(args_list, kwargs_list):
            session_id, data_key, obj, level = self.put.bind(*args, **kwargs)
            if level is None:
                level = self._get_default_level(obj)
            size = await asyncio.to_thread(calc_data_size, obj)
            if last_level is not None:
                assert last_level == level
            last_level = level
            objs.append(obj)
            data_keys.append(data_key)
            sizes.append(size)

        await self.request_quota_with_spill(level, sum(sizes))

        data_infos = []
        put_infos = []
        quota_delta = 0
        for size, data_key, obj in zip(sizes, data_keys, objs):
            object_info = await self._clients[level].put(obj)
            data_info = build_data_info(object_info, level, size, self._band_name)
            data_infos.append(data_info)
            if (
                object_info.size is not None
                and data_info.memory_size != object_info.size
            ):
                # we request memory size before putting, when put finishes,
                # update quota to the true store size
                quota_delta += object_info.size - data_info.memory_size
            put_infos.append(
                self._data_manager_ref.put_data_info.delay(
                    session_id, data_key, data_info, object_info
                )
            )
        await self._quota_refs[level].update_quota(quota_delta)
        await self._data_manager_ref.put_data_info.batch(*put_infos)
        await self.notify_spillable_space(level)
        return data_infos

    def _get_data_infos_arg(self, session_id: str, data_key: str, error: str = "raise"):
        infos = self._data_manager_ref.get_data_infos.delay(
            session_id, data_key, self._band_name, error
        )
        return infos, session_id, data_key

    async def delete_object(
        self,
        session_id: str,
        data_key: Any,
        data_size: Union[int, float],
        object_id: Any,
        level: StorageLevel,
    ):
        await self._data_manager_ref.delete_data_info(
            session_id, data_key, level, self._band_name
        )
        await self._clients[level].delete(object_id)
        await self._quota_refs[level].release_quota(data_size)

    @mo.extensible
    async def delete(self, session_id: str, data_key: str, error: str = "raise"):
        if error not in ("raise", "ignore"):  # pragma: no cover
            raise ValueError("error must be raise or ignore")

        all_infos = await self._data_manager_ref.get_data_infos(
            session_id, data_key, self._band_name, error
        )
        if not all_infos:
            return

        key_to_infos = (
            all_infos if isinstance(all_infos, dict) else {data_key: all_infos}
        )

        for key, infos in key_to_infos.items():
            for info in infos:
                level = info.level
                await self._data_manager_ref.delete_data_info(
                    session_id, key, level, self._band_name
                )
                await self._clients[level].delete(info.object_id)
                await self._quota_refs[level].release_quota(info.store_size)

    @delete.batch
    async def batch_delete(self, args_list, kwargs_list):
        get_infos = []
        session_id = None
        data_keys = []
        for args, kwargs in zip(args_list, kwargs_list):
            infos, session_id, data_key = self._get_data_infos_arg(*args, **kwargs)
            get_infos.append(infos)
            data_keys.append(data_key)
        infos_list = await self._data_manager_ref.get_data_infos.batch(*get_infos)

        delete_infos = []
        to_removes = []
        level_sizes = defaultdict(lambda: 0)
        for all_infos, data_key in zip(infos_list, data_keys):
            if not all_infos:
                # data not exist and error == 'ignore'
                continue
            key_to_infos = (
                all_infos if isinstance(all_infos, dict) else {data_key: all_infos}
            )

            for key, infos in key_to_infos.items():
                for info in infos:
                    level = info.level
                    delete_infos.append(
                        self._data_manager_ref.delete_data_info.delay(
                            session_id, key, level, info.band
                        )
                    )
                    to_removes.append((level, info.object_id))
                    level_sizes[level] += info.store_size

        if not delete_infos:
            # no data to remove
            return

        await self._data_manager_ref.delete_data_info.batch(*delete_infos)
        await asyncio.gather(
            *[self._clients[level].delete(object_id) for level, object_id in to_removes]
        )
        for level, size in level_sizes.items():
            await self._quota_refs[level].release_quota(size)

    @mo.extensible
    async def open_reader(self, session_id: str, data_key: str) -> StorageFileObject:
        data_info = await self._data_manager_ref.get_data_info(
            session_id, data_key, self._band_name
        )
        reader = await self._clients[data_info.level].open_reader(data_info.object_id)
        return reader

    @open_reader.batch
    async def batch_open_readers(self, args_list, kwargs_list):
        get_data_infos = []
        for args, kwargs in zip(args_list, kwargs_list):
            get_data_infos.append(
                self._data_manager_ref.get_data_info.delay(
                    *args, band_name=self._band_name, **kwargs
                )
            )
        data_infos = await self._data_manager_ref.get_data_info.batch(*get_data_infos)
        return await asyncio.gather(
            *[
                self._clients[data_info.level].open_reader(data_info.object_id)
                for data_info in data_infos
            ]
        )

    @mo.extensible
    async def open_writer(
        self,
        session_id: str,
        data_key: str,
        size: int,
        level: StorageLevel,
        request_quota=True,
    ) -> WrappedStorageFileObject:
        if request_quota:
            await self.request_quota_with_spill(level, size)
        writer = await self._clients[level].open_writer(size)
        return WrappedStorageFileObject(
            writer,
            level,
            size,
            session_id,
            data_key,
            self._data_manager_ref,
            self._clients[level],
        )

    @open_writer.batch
    async def batch_open_writers(self, args_list, kwargs_list):
        extracted_args = None
        data_keys, sizes = [], []
        for args, kwargs in zip(args_list, kwargs_list):
            session_id, data_key, size, level, request_quota = self.open_writer.bind(
                *args, **kwargs
            )
            if extracted_args:
                assert extracted_args == (session_id, level, request_quota)
            extracted_args = (session_id, level, request_quota)
            data_keys.append(data_key)
            sizes.append(size)
        session_id, level, request_quota = extracted_args
        if request_quota:  # pragma: no cover
            await self.request_quota_with_spill(level, sum(sizes))
        writers = await asyncio.gather(
            *[self._clients[level].open_writer(size) for size in sizes]
        )
        wrapped_writers = []
        for writer, size, data_key in zip(writers, sizes, data_keys):
            wrapped_writers.append(
                WrappedStorageFileObject(
                    writer,
                    level,
                    size,
                    session_id,
                    data_key,
                    self._data_manager_ref,
                    self._clients[level],
                )
            )
        return wrapped_writers

    async def _get_meta_api(self, session_id: str):
        if self._supervisor_address is None:
            cluster_api = await ClusterAPI.create(self.address)
            [self._supervisor_address] = await cluster_api.get_supervisors_by_keys(
                [session_id]
            )

        return await MetaAPI.create(
            session_id=session_id, address=self._supervisor_address
        )

    async def _fetch_remote(
        self,
        session_id: str,
        data_keys: List[Union[str, tuple]],
        remote_band: BandType,
        error: str,
    ):
        remote_manager_ref: mo.ActorRefType[DataManagerActor] = await mo.actor_ref(
            uid=DataManagerActor.default_uid(), address=remote_band[0]
        )
        get_data_infos = []
        for data_key in data_keys:
            get_data_infos.append(
                remote_manager_ref.get_data_info.delay(session_id, data_key, error)
            )
        data_infos = await remote_manager_ref.get_data_info.batch(*get_data_infos)
        data_infos, data_keys = zip(
            *[
                (data_info, data_key)
                for data_info, data_key in zip(data_infos, data_keys)
                if data_info is not None
            ]
        )
        put_data_info_delays = []
        fetch_tasks = []
        for data_info, data_key in zip(data_infos, data_keys):
            put_data_info_delays.append(
                self._data_manager_ref.put_data_info.delay(
                    session_id, data_key, data_info, None
                )
            )
            fetch_tasks.append(
                self._clients[StorageLevel.REMOTE].fetch(data_info.object_id)
            )
        await self._data_manager_ref.put_data_info.batch(*put_data_info_delays)
        await asyncio.gather(*fetch_tasks)

    async def _fetch_via_transfer(
        self,
        session_id: str,
        data_keys: List[Union[str, tuple]],
        level: StorageLevel,
        remote_band: BandType,
        fetch_band_name: str,
        error: str,
    ):
        from .transfer import SenderManagerActor

        logger.debug("Begin to fetch %s from band %s", data_keys, remote_band)
        sender_ref: mo.ActorRefType[SenderManagerActor] = await mo.actor_ref(
            address=remote_band[0], uid=SenderManagerActor.gen_uid(remote_band[1])
        )
        await sender_ref.send_batch_data(
            session_id,
            data_keys,
            self._data_manager_ref.address,
            level,
            fetch_band_name,
            error=error,
        )
        logger.debug("Finish fetching %s from band %s", data_keys, remote_band)

    async def fetch_batch(
        self,
        session_id: str,
        data_keys: List[str],
        level: StorageLevel,
        band_name: str,
        address: str,
        error: str,
    ):
        if error not in ("raise", "ignore"):  # pragma: no cover
            raise ValueError("error must be raise or ignore")

        meta_api = await self._get_meta_api(session_id)
        remote_keys = defaultdict(set)
        missing_keys = []
        get_metas = []
        get_info_delays = []
        for data_key in data_keys:
            get_info_delays.append(
                self._data_manager_ref.get_data_info.delay(
                    session_id, data_key, band_name, error="ignore"
                )
            )
        data_infos = await self._data_manager_ref.get_data_info.batch(*get_info_delays)
        pin_delays = []
        for data_key, info in zip(data_keys, data_infos):
            # for gpu bands, need transfer between gpu cards
            if info is not None:
                if band_name and band_name != info.band:
                    missing_keys.append(data_key)
                else:
                    pin_delays.append(
                        self._data_manager_ref.pin.delay(
                            session_id, data_key, self._band_name
                        )
                    )
            else:
                # Not exists in local, fetch from remote worker
                missing_keys.append(data_key)
        if address is None or band_name is None:
            # some mapper keys are absent, specify error='ignore'
            # remember that meta only records those main keys
            get_metas = [
                (
                    meta_api.get_chunk_meta.delay(
                        data_key[0] if isinstance(data_key, tuple) else data_key,
                        fields=["bands"],
                        error="ignore",
                    )
                )
                for data_key in missing_keys
            ]
        await self._data_manager_ref.pin.batch(*pin_delays)

        if get_metas:
            metas = await meta_api.get_chunk_meta.batch(*get_metas)
        else:  # pragma: no cover
            metas = [{"bands": [(address, band_name)]}] * len(missing_keys)
        assert len(metas) == len(missing_keys)
        for data_key, bands in zip(missing_keys, metas):
            if bands is not None:
                remote_keys[bands["bands"][0]].add(data_key)
        transfer_tasks = []
        fetch_keys = []
        for band, keys in remote_keys.items():
            if StorageLevel.REMOTE in self._quota_refs:
                # if storage support remote level, just fetch object id
                transfer_tasks.append(
                    self._fetch_remote(session_id, list(keys), band, error)
                )
            else:
                # fetch via transfer
                transfer_tasks.append(
                    self._fetch_via_transfer(
                        session_id, list(keys), level, band, band_name or band[1], error
                    )
                )
            fetch_keys.extend(list(keys))

        await asyncio.gather(*transfer_tasks)

        append_bands_delays = []
        for data_key in fetch_keys:
            # skip shuffle keys
            if isinstance(data_key, tuple):
                continue
            append_bands_delays.append(
                meta_api.add_chunk_bands.delay(
                    data_key,
                    [(self.address, self._band_name)],
                )
            )
        if append_bands_delays:
            await meta_api.add_chunk_bands.batch(*append_bands_delays)
        return fetch_keys

    async def request_quota_with_spill(self, level: StorageLevel, size: int):
        if await self._quota_refs[level].request_quota(size):
            return
        else:
            total, used = await self._quota_refs[level].get_quota()
            await self.spill(level, int(used + size - total), size)
            await self._quota_refs[level].request_quota(size)
            logger.debug(
                "Spill is triggered, request %s bytes of %s finished", size, level
            )

    async def notify_spillable_space(self, level):
        if await self._spill_manager_refs[level].has_spill_task():
            total, used = await self._quota_refs[level].get_quota()
            tasks = []
            if total is not None:
                spillable_size = await self._data_manager_ref.get_spillable_size(
                    level, self._band_name
                )
                tasks.append(
                    self._spill_manager_refs[level].notify_spillable_space(
                        spillable_size, total - used
                    )
                )
            await asyncio.gather(*tasks)

    async def spill(self, level: StorageLevel, request_size: int, object_size: int):
        from .spill import spill

        try:
            await spill(
                request_size, level, self._band_name, self._data_manager_ref, self
            )
        except NoDataToSpill:
            logger.warning(
                "No data to spill %s bytes, waiting more space", request_size
            )
            size = await self._spill_manager_refs[level].wait_for_space(object_size)
            await spill(size, level, self._band_name, self._data_manager_ref, self)

    async def list(self, level: StorageLevel) -> List:
        return await self._data_manager_ref.list(level, self._band_name)

    @mo.extensible
    async def unpin(self, session_id: str, data_key: str, error: str = "raise"):
        levels = await self._data_manager_ref.unpin(
            session_id, [data_key], self._band_name, error
        )
        if levels:
            await self.notify_spillable_space(levels[0])

    @unpin.batch
    async def batch_unpin(self, args_list, kwargs_list):
        extracted_args = []
        data_keys = []
        for args, kw in zip(args_list, kwargs_list):
            session_id, data_key, error = self.unpin.bind(*args, **kw)
            if extracted_args:
                assert extracted_args == (session_id, error)
            extracted_args = session_id, error
            data_keys.append(data_key)
        if extracted_args:
            session_id, error = extracted_args
            levels = await self._data_manager_ref.unpin(
                session_id, data_keys, self._band_name, error
            )
            for level in levels:
                await self.notify_spillable_space(level)

    async def get_storage_level_info(self, level: StorageLevel) -> StorageInfo:
        quota_ref = self._quota_refs[level]
        total_size, used_size = await quota_ref.get_quota()
        return StorageInfo(
            storage_level=level,
            total_size=int(total_size) if total_size else total_size,
            used_size=int(used_size),
        )

    async def get_storage_backend_info(self, level: StorageLevel) -> dict:
        return self._clients[level].backend_info



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import scipy.special as spspecial

from ..utils import infer_dtype, implement_scipy
from .core import _register_special_op, TensorSpecialMultiOp


@_register_special_op
class TensorEllipHarm(TensorSpecialMultiOp):
    _ARG_COUNT = 5
    _func_name = "ellip_harm"


@implement_scipy(spspecial.ellip_harm)
@infer_dtype(spspecial.ellip_harm)
def ellip_harm(h2, k2, n, p, s, signm=1, signn=1, **kwargs):
    op = TensorEllipHarm(**kwargs)
    return op(h2, k2, n, p, s, signm, signn)


@_register_special_op
class TensorEllipHarm2(TensorSpecialMultiOp):
    _ARG_COUNT = 5
    _func_name = "ellip_harm_2"


@implement_scipy(spspecial.ellip_harm_2)
@infer_dtype(spspecial.ellip_harm_2)
def ellip_harm_2(h2, k2, n, p, s, **kwargs):
    op = TensorEllipHarm2(**kwargs)
    return op(h2, k2, n, p, s)


@_register_special_op
class TensorEllipNormal(TensorSpecialMultiOp):
    _ARG_COUNT = 4
    _func_name = "ellip_normal"


@implement_scipy(spspecial.ellip_normal)
@infer_dtype(spspecial.ellip_normal)
def ellip_normal(h2, k2, n, p, **kwargs):
    op = TensorEllipNormal(**kwargs)
    return op(h2, k2, n, p)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union, Callable

import numpy as np

from ....lib.aio import alru_cache
from ....utils import deserialize_serializable, serialize_serializable
from ...web import web_api, MarsServiceWebAPIHandler, MarsWebAPIClientMixin
from .core import AbstractMutableAPI


class MutableWebAPIHandler(MarsServiceWebAPIHandler):
    _root_pattern = "/api/session/(?P<session_id>[^/]+)/mutable"

    @alru_cache(cache_exceptions=False)
    async def _get_cluster_api(self):
        from ...cluster import ClusterAPI

        return await ClusterAPI.create(self._supervisor_addr)

    @alru_cache(cache_exceptions=False)
    async def _get_oscar_mutable_api(self, session_id: str):
        from .oscar import MutableAPI

        cluster_api = await self._get_cluster_api()
        [address] = await cluster_api.get_supervisors_by_keys([session_id])
        return await MutableAPI.create(session_id, address)

    @web_api("", method="post")
    async def create_mutable_tensor(self, session_id: str):
        body_args = (
            deserialize_serializable(self.request.body) if self.request.body else None
        )
        shape = body_args.get("shape")
        dtype = body_args.get("dtype")
        name = body_args.get("name")
        default_value = body_args.get("default_value")
        chunk_size = body_args.get("chunk_size")

        oscar_api = await self._get_oscar_mutable_api(session_id)
        res = await oscar_api.create_mutable_tensor(
            shape, dtype, name, default_value, chunk_size
        )
        self.write(serialize_serializable(res))

    @web_api("(?P<name>[^/]+)", method="get")
    async def get_mutable_tensor(self, session_id: str, name: str):
        oscar_api = await self._get_oscar_mutable_api(session_id)
        res = await oscar_api.get_mutable_tensor(name)
        self.write(serialize_serializable(res))

    @web_api("(?P<name>[^/]+)/seal", method="post")
    async def seal_mutable_tensor(self, session_id: str, name: str):  # pragma: no cover
        body_args = (
            deserialize_serializable(self.request.body) if self.request.body else None
        )
        timestamp = body_args.get("timestamp")

        oscar_api = await self._get_oscar_mutable_api(session_id)
        res = await oscar_api.seal_mutable_tensor(name, timestamp)
        self.write(serialize_serializable(res))

    @web_api("(?P<name>[^/]+)/read", method="post")
    async def read_mutable(self, session_id: str, name: str):  # pragma: no cover
        body_args = (
            deserialize_serializable(self.request.body) if self.request.body else None
        )
        index = body_args.get("index")
        timestamp = body_args.get("timestamp")

        oscar_api = await self._get_oscar_mutable_api(session_id)
        res = await oscar_api.read(name, index, timestamp)
        self.write(serialize_serializable(res))

    @web_api("(?P<name>[^/]+)/write", method="post")
    async def write_mutable(self, session_id: str, name: str):  # pragma: no cover
        body_args = (
            deserialize_serializable(self.request.body) if self.request.body else None
        )
        index = body_args.get("index")
        value = body_args.get("value")
        timestamp = body_args.get("timestamp")

        oscar_api = await self._get_oscar_mutable_api(session_id)
        res = await oscar_api.write(name, index, value, timestamp)
        self.write(serialize_serializable(res))


web_handlers = {
    MutableWebAPIHandler.get_root_pattern(): MutableWebAPIHandler,
}


class WebMutableAPI(AbstractMutableAPI, MarsWebAPIClientMixin):
    def __init__(
        self, session_id: str, address: str, request_rewriter: Callable = None
    ):
        self._session_id = session_id
        self._address = address.rstrip("/")
        self.request_rewriter = request_rewriter

    async def create_mutable_tensor(
        self,
        shape: tuple,
        dtype: Union[np.dtype, str],
        name: str = None,
        default_value: Union[int, float] = 0,
        chunk_size: Union[tuple, int] = None,
    ):
        path = f"{self._address}/api/session/{self._session_id}/mutable"
        params = dict(
            shape=shape,
            dtype=dtype,
            name=name,
            default_value=default_value,
            chunk_size=chunk_size,
        )
        body = serialize_serializable(params)
        res = await self._request_url(
            path=path,
            method="POST",
            data=body,
            headers={"Content-Type": "application/octet-stream"},
        )
        return deserialize_serializable(res.body)

    async def get_mutable_tensor(self, name: str):
        path = f"{self._address}/api/session/{self._session_id}/mutable/{name}"
        res = await self._request_url(
            path=path,
            method="GET",
            headers={"Content-Type": "application/octet-stream"},
        )
        return deserialize_serializable(res.body)

    async def seal_mutable_tensor(self, name: str, timestamp=None):
        path = f"{self._address}/api/session/{self._session_id}/mutable/{name}/seal"
        params = dict(timestamp=timestamp)
        body = serialize_serializable(params)
        res = await self._request_url(
            path=path,
            method="POST",
            data=body,
            headers={"Content-Type": "application/octet-stream"},
            request_timeout=3600,
        )
        return deserialize_serializable(res.body)

    async def read(self, name: str, index, timestamp=None):
        path = f"{self._address}/api/session/{self._session_id}/mutable/{name}/read"
        params = dict(index=index, timestamp=timestamp)
        body = serialize_serializable(params)
        res = await self._request_url(
            path=path,
            method="POST",
            data=body,
            headers={"Content-Type": "application/octet-stream"},
            request_timeout=3600,
        )
        return deserialize_serializable(res.body)

    async def write(self, name: str, index, value, timestamp=None):
        path = f"{self._address}/api/session/{self._session_id}/mutable/{name}/write"
        params = dict(index=index, value=value, timestamp=timestamp)
        body = serialize_serializable(params)
        res = await self._request_url(
            path=path,
            method="POST",
            data=body,
            headers={"Content-Type": "application/octet-stream"},
            request_timeout=3600,
        )
        return deserialize_serializable(res.body)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta, abstractmethod
from typing import Any, List

from ....core import OperandType, TileableType, CHUNK_TYPE
from ....dataframe.datasource.core import ColumnPruneSupportedDataSourceMixin
from ....dataframe.utils import parse_index
from ....utils import implements
from ..core import OptimizationRule, OptimizationRecord, OptimizationRecordType


class PruneDataSource(OptimizationRule, metaclass=ABCMeta):
    def _all_successor_prune_pushdown(self, successors: List[TileableType]):
        for succ in successors:
            rule_types = self._optimizer_cls.get_rule_types(type(succ.op))
            if rule_types is None:
                return False

            prune_rule_types = [
                rule_type
                for rule_type in rule_types
                if issubclass(rule_type, PruneDataSource)
            ]
            if not prune_rule_types:
                return False

            for rule_type in prune_rule_types:
                rule = rule_type(self._graph, self._records, self._optimizer_cls)
                if not rule._need_prune(succ.op):
                    return False
        return True

    @implements(OptimizationRule.match)
    def match(self, op: OperandType) -> bool:
        node = op.outputs[0]
        input_node = self._graph.predecessors(node)[0]
        successors = self._graph.successors(input_node)
        return self._all_successor_prune_pushdown(successors)

    @abstractmethod
    def _need_prune(self, op: OperandType) -> bool:
        """
        Check if this operand can prune

        Returns
        -------
        need_prune : bool
        """

    @abstractmethod
    def _get_selected_columns(self, op: OperandType) -> List[Any]:
        """
        Get selected columns to prune data source.

        Parameters
        ----------
        op : OperandType
            Operand.

        Returns
        -------
        columns : list
            Columns selected.
        """

    def _merge_selected_columns(self, selected_columns: List[Any], op: OperandType):
        input_node = self._graph.predecessors(op.outputs[0])[0]
        original_node = self._records.get_original_entity(input_node)
        if original_node is None:
            # not pruned before
            original_all_columns = input_node.dtypes.index.tolist()
            if set(selected_columns) != set(original_all_columns):
                # not prune all fields
                return [c for c in original_all_columns if c in selected_columns]
            else:
                return []
        else:
            # pruned before
            original_all_columns = original_node.dtypes.index.tolist()
            original_pruned_columns = input_node.op.get_columns()
            pruned_columns_set = set(selected_columns) | set(original_pruned_columns)
            # pruned before, cannot revert it,
            # so we just return pruned columns
            # even though no columns pruned
            return [c for c in original_all_columns if c in pruned_columns_set]

    @implements(OptimizationRule.apply)
    def apply(self, op: OperandType):
        node = op.outputs[0]
        data_source_node = self._graph.predecessors(node)[0]

        if (
            isinstance(node, CHUNK_TYPE)
            and self._graph.count_successors(data_source_node) == 1
        ):
            # merge into data source only for chunk
            data_source_params = node.params.copy()
            data_source_params.update(data_source_node.extra_params)
            data_source_op = data_source_node.op.copy()
            data_source_op._key = data_source_node.op.key
            data_source_op._output_types = op.output_types
            if node.ndim == 1:
                data_source_op.set_pruned_columns(node.name, keep_order=True)
            else:
                data_source_op.set_pruned_columns(
                    node.dtypes.index.tolist(), keep_order=True
                )
            new_entity = (
                data_source_op.new_tileable
                if not isinstance(node, CHUNK_TYPE)
                else data_source_op.new_chunk
            )
            new_data_source_node = new_entity(
                data_source_node.inputs, kws=[data_source_params]
            ).data
            new_data_source_node._key = node.key
            new_data_source_node._id = node.id
            # just remove the input data
            self._graph.add_node(new_data_source_node)
            for succ in self._graph.successors(node):
                self._graph.add_edge(new_data_source_node, succ)
            self._graph.remove_node(data_source_node)
            self._graph.remove_node(node)

            # mark optimization record
            # the input node is removed
            self._records.append_record(
                OptimizationRecord(
                    data_source_node, None, OptimizationRecordType.delete
                )
            )
            self._records.append_record(
                OptimizationRecord(
                    node, new_data_source_node, OptimizationRecordType.replace
                )
            )
            new_outputs = [new_data_source_node]
        else:
            selected_columns: List[Any] = self._get_selected_columns(op)
            original_node = self._records.get_original_entity(data_source_node)
            if original_node is not None:
                # pruned before
                dtypes = original_node.dtypes
            else:
                dtypes = data_source_node.dtypes
            data_source_params = data_source_node.params.copy()
            data_source_params["shape"] = (
                data_source_node.shape[0],
                len(selected_columns),
            )
            data_source_params["dtypes"] = dtypes = dtypes[selected_columns]
            data_source_params["columns_value"] = parse_index(
                dtypes.index, store_data=True
            )
            data_source_params.update(data_source_node.extra_params)
            data_source_node_op = data_source_node.op.copy()
            data_source_node_op._key = data_source_node.op.key
            data_source_node_op.set_pruned_columns(selected_columns, keep_order=True)
            new_data_source_node = data_source_node_op.new_tileable(
                data_source_node_op.inputs, kws=[data_source_params]
            ).data

            self._replace_node(data_source_node, new_data_source_node)
            # mark optimization record
            self._records.append_record(
                OptimizationRecord(
                    data_source_node,
                    new_data_source_node,
                    OptimizationRecordType.replace,
                )
            )

            new_op = op.copy()
            new_op._key = op.key
            kws = []
            for out in op.outputs:
                params = out.params.copy()
                params.update(out.extra_params)
                kws.append(params)
            new_entity = (
                new_op.new_tileables
                if not isinstance(node, CHUNK_TYPE)
                else new_op.new_chunks
            )
            new_outputs = [t.data for t in new_entity([new_data_source_node], kws=kws)]

            for out, new_out in zip(op.outputs, new_outputs):
                new_out._id = out.id
                new_out._key = out.key
                self._replace_node(out, new_out)
                # mark optimization record
                self._records.append_record(
                    OptimizationRecord(out, new_out, OptimizationRecordType.replace)
                )

        for out, new_out in zip(op.outputs, new_outputs):
            # check out if it's in result
            try:
                i = self._graph.results.index(out)
                self._graph.results[i] = new_out
            except ValueError:
                pass


class GetitemPruneDataSource(PruneDataSource):
    def _need_prune(self, op: OperandType) -> bool:
        data_source_node = self._graph.predecessors(op.outputs[0])[0]
        input_can_be_pruned = isinstance(
            data_source_node.op, ColumnPruneSupportedDataSourceMixin
        )
        if (
            input_can_be_pruned
            and data_source_node not in self._graph.results
            and op.col_names is not None
        ):
            selected_columns = self._get_selected_columns(op)
            if not isinstance(op.outputs[0], CHUNK_TYPE) and not selected_columns:
                # no columns selected, skip
                return False
            return True
        return False

    def _get_selected_columns(self, op: OperandType) -> List[str]:
        columns = op.col_names if isinstance(op.col_names, list) else [op.col_names]
        return self._merge_selected_columns(columns, op)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pickle  # nosec  # pylint: disable=import_pickle

from .train import StatsModelsTrain
from .predict import StatsModelsPredict

try:
    from statsmodels.base.distributed_estimation import DistributedModel
except ImportError:
    DistributedModel = None


class MarsDistributedModel:
    def __init__(
        self,
        factor=None,
        num_partitions=None,
        model_class=None,
        init_kwds=None,
        estimation_method=None,
        estimation_kwds=None,
        join_method=None,
        join_kwds=None,
        results_class=None,
        results_kwds=None,
    ):
        self._factor = factor
        self._sm_model = DistributedModel(
            num_partitions or 10,
            model_class=model_class,
            init_kwds=init_kwds,
            estimation_method=estimation_method,
            estimation_kwds=estimation_kwds,
            join_method=join_method,
            join_kwds=join_kwds,
            results_class=results_class,
            results_kwds=results_kwds,
        )

    def fit(self, endog, exog, session=None, **kwargs):
        num_partitions = None if self._factor is not None else self._sm_model.partitions
        run_kwargs = kwargs.pop("run_kwargs", dict())
        op = StatsModelsTrain(
            endog=endog,
            exog=exog,
            num_partitions=num_partitions,
            factor=self._factor,
            model_class=self._sm_model.model_class,
            init_kwds=self._sm_model.init_kwds,
            fit_kwds=kwargs,
            estimation_method=self._sm_model.estimation_method,
            estimation_kwds=self._sm_model.estimation_kwds,
            join_method=self._sm_model.join_method,
            join_kwds=self._sm_model.join_kwds,
            results_class=self._sm_model.results_class,
            results_kwds=self._sm_model.results_kwds,
        )
        result = (
            op(exog, endog)
            .execute(session=session, **run_kwargs)
            .fetch(session=session)
        )
        return MarsResults(pickle.loads(result))  # nosec


class MarsResults:
    def __init__(self, model):
        self._model = model

    @property
    def model(self):
        return self._model

    def __getattr__(self, item):
        if item == "_model":
            raise AttributeError
        return getattr(self._model, item)

    def __mars_tokenize__(self):
        return pickle.dumps(self.model)

    def predict(self, exog, *args, **kwargs):
        session = kwargs.pop("session", None)
        run_kwargs = kwargs.pop("run_kwargs", dict())
        op = StatsModelsPredict(
            model_results=self, predict_args=args, predict_kwargs=kwargs
        )
        return op(exog).execute(session=session, **run_kwargs)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from urllib.parse import urlparse
from typing import Any, Dict, Type, Tuple
from numbers import Number
from collections import defaultdict

from .backend import get_backend
from .context import get_context
from .core import _Actor, _StatelessActor, ActorRef


async def create_actor(actor_cls, *args, uid=None, address=None, **kwargs) -> ActorRef:
    ctx = get_context()
    return await ctx.create_actor(actor_cls, *args, uid=uid, address=address, **kwargs)


async def has_actor(actor_ref: ActorRef) -> bool:
    ctx = get_context()
    return await ctx.has_actor(actor_ref)


async def destroy_actor(actor_ref: ActorRef):
    ctx = get_context()
    return await ctx.destroy_actor(actor_ref)


async def actor_ref(*args, **kwargs) -> ActorRef:
    ctx = get_context()
    return await ctx.actor_ref(*args, **kwargs)


async def kill_actor(actor_ref):
    ctx = get_context()
    return await ctx.kill_actor(actor_ref)


async def create_actor_pool(address: str, n_process: int = None, **kwargs):
    if address is None:
        raise ValueError("address has to be provided")
    if "://" not in address:
        scheme = None
    else:
        scheme = urlparse(address).scheme or None

    return await get_backend(scheme).create_actor_pool(
        address, n_process=n_process, **kwargs
    )


async def wait_actor_pool_recovered(address: str, main_pool_address: str = None):
    ctx = get_context()
    return await ctx.wait_actor_pool_recovered(address, main_pool_address)


async def get_pool_config(address: str):
    ctx = get_context()
    return await ctx.get_pool_config(address)


def setup_cluster(address_to_resources: Dict[str, Dict[str, Number]]):
    scheme_to_address_resources = defaultdict(dict)
    for address, resources in address_to_resources.items():
        if address is None:
            raise ValueError("address has to be provided")
        if "://" not in address:
            scheme = None
        else:
            scheme = urlparse(address).scheme or None

        scheme_to_address_resources[scheme][address] = resources
    for scheme, address_resources in scheme_to_address_resources.items():
        get_backend(scheme).get_driver_cls().setup_cluster(address_resources)


class AsyncActorMixin:
    @classmethod
    def default_uid(cls):
        return cls.__name__

    def __new__(cls, *args, **kwargs):
        try:
            return _actor_implementation[cls](*args, **kwargs)
        except KeyError:
            return super().__new__(cls, *args, **kwargs)

    async def __post_create__(self):
        """
        Method called after actor creation
        """
        return await super().__post_create__()

    async def __pre_destroy__(self):
        """
        Method called before actor destroy
        """
        return await super().__pre_destroy__()

    async def __on_receive__(self, message: Tuple[Any]):
        """
        Handle message from other actors and dispatch them to user methods

        Parameters
        ----------
        message : tuple
            Message shall be (method_name,) + args + (kwargs,)
        """
        return await super().__on_receive__(message)


class Actor(AsyncActorMixin, _Actor):
    pass


class StatelessActor(AsyncActorMixin, _StatelessActor):
    pass


_actor_implementation: Dict[Type[Actor], Type[Actor]] = dict()


def register_actor_implementation(actor_cls: Type[Actor], impl_cls: Type[Actor]):
    _actor_implementation[actor_cls] = impl_cls


def unregister_actor_implementation(actor_cls: Type[Actor]):
    try:
        del _actor_implementation[actor_cls]
    except KeyError:
        pass



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import enum
import functools
import time
from dataclasses import dataclass, field
from typing import (
    AsyncGenerator,
    Awaitable,
    Callable,
    Dict,
    Optional,
    Set,
    Tuple,
    TypeVar,
)

from ...serialization.serializables import (
    Serializable,
    Float64Field,
    Int32Field,
    Int64Field,
    StringField,
)
from ...storage import StorageLevel
from ..core import NodeRole


class NodeStatus(enum.Enum):
    STARTING = 0
    READY = 1
    DEGENERATED = 2
    STOPPING = 3
    STOPPED = -1


@dataclass
class NodeInfo:
    role: NodeRole
    status: NodeStatus = NodeStatus.READY
    update_time: float = field(default_factory=time.time)
    env: Dict = field(default_factory=dict)
    resource: Dict = field(default_factory=dict)
    detail: Dict = field(default_factory=dict)


class WatchNotifier:
    _events: Set[asyncio.Event]

    def __init__(self):
        self._event = asyncio.Event()
        self._lock = asyncio.Lock()
        self._version = 0

    async def watch(self, version: Optional[int] = None):
        if version != self._version:
            return self._version
        await self._event.wait()
        return self._version

    async def notify(self):
        async with self._lock:
            self._version += 1
            self._event.set()
            self._event = asyncio.Event()


RetType = TypeVar("RetType")


def watch_method(
    func: Callable[..., Awaitable[Tuple[int, RetType]]]
) -> Callable[..., AsyncGenerator[RetType, None]]:
    @functools.wraps(func)
    async def wrapped(*args, **kwargs):
        if "version" in kwargs:
            yield await func(*args, **kwargs)
            return

        kwargs["version"] = None
        while True:
            version, val = await func(*args, **kwargs)
            kwargs["version"] = version
            yield val

    return wrapped


class WorkerSlotInfo(Serializable):
    slot_id: int = Int32Field("slot_id")
    session_id: str = StringField("session_id")
    subtask_id: str = StringField("subtask_id")
    processor_usage: float = Float64Field("processor_usage")


class QuotaInfo(Serializable):
    quota_size: int = Int64Field("quota_size")
    allocated_size: int = Int64Field("allocated_size")
    hold_size: int = Int64Field("hold_size")


class StorageInfo(Serializable):
    storage_level: StorageLevel = Int32Field(
        "storage_level", on_serialize=lambda x: x.value, on_deserialize=StorageLevel
    )
    total_size: int = Int64Field("total_size")
    used_size: int = Int64Field("used_size")
    pinned_size: int = Int64Field("pinned_size", default=None)


class DiskInfo(Serializable):
    path: str = StringField("path")
    limit_size: int = Int64Field("limit_size", default=None)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from threading import Thread

from .... import opcodes as OperandDef
from ....core import NotSupportTile
from ....serialization.serializables import Int32Field
from ....utils import to_binary
from ...operands import LearnOperand, LearnOperandMixin, OutputType


class StartTracker(LearnOperand, LearnOperandMixin):
    _op_type_ = OperandDef.START_TRACKER
    _op_module_ = "learn.contrib.xgboost"

    n_workers = Int32Field("n_workers", default=None)

    def __init__(self, output_types=None, pure_depends=None, **kw):
        super().__init__(
            _output_types=output_types,
            _pure_depends=pure_depends,
            **kw,
        )
        if self.output_types is None:
            self.output_types = [OutputType.object]

    @classmethod
    def tile(cls, op):
        raise NotSupportTile("StartTracker is a chunk op")

    @classmethod
    def execute(cls, ctx, op):
        """Start Rabit tracker"""
        from .tracker import RabitTracker

        env = {"DMLC_NUM_WORKER": op.n_workers}
        rabit_context = RabitTracker(
            host_ip=ctx.get_local_host_ip(), n_workers=op.n_workers
        )
        env.update(rabit_context.worker_envs())

        rabit_context.start(op.n_workers)
        thread = Thread(target=rabit_context.join)
        thread.daemon = True
        thread.start()

        rabit_args = [to_binary(f"{k}={v}") for k, v in env.items()]
        ctx[op.outputs[0].key] = rabit_args



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, Union

import numpy as np
from sklearn.base import (
    MetaEstimatorMixin,
    BaseEstimator as SklearnBaseEstimator,
    RegressorMixin as SklearnRegressorMixin,
    ClassifierMixin as SklearnClassifierMixin,
)

from .. import remote as mr
from .. import tensor as mt
from ..core import ENTITY_TYPE
from .base import BaseEstimator, RegressorMixin, ClassifierMixin
from .metrics import get_scorer
from .utils import copy_learned_attributes, check_array


def _wrap(estimator: SklearnBaseEstimator, method, X, y, **kwargs):
    X = X.fetch() if isinstance(X, ENTITY_TYPE) else X
    y = y.fetch() if isinstance(y, ENTITY_TYPE) else y
    return getattr(estimator, method)(X, y, **kwargs)


class ParallelPostFit(BaseEstimator, MetaEstimatorMixin):
    """
    Meta-estimator for parallel predict and transform.

    Parameters
    ----------
    estimator : Estimator
        The underlying estimator that is fit.

    scoring : string or callable, optional
        A single string (see :ref:`scoring_parameter`) or a callable
        (see :ref:`scoring`) to evaluate the predictions on the test set.

        For evaluating multiple metrics, either give a list of (unique)
        strings or a dict with names as keys and callables as values.

        NOTE that when using custom scorers, each scorer should return a
        single value. Metric functions returning a list/array of values
        can be wrapped into multiple scorers that return one value each.

        See :ref:`multimetric_grid_search` for an example.

        .. warning::

           If None, the estimator's default scorer (if available) is used.
           Most scikit-learn estimators will convert large Mars tensors to
           a single NumPy array, which may exhaust the memory of your worker.
           You probably want to always specify `scoring`.

    Notes
    -----

    .. warning::

       This class is not appropriate for parallel or distributed *training*
       on large datasets. For that, see :class:`Incremental`, which provides
       distributed (but sequential) training. If you're doing distributed
       hyperparameter optimization on larger-than-memory datasets, see
       :class:`mars.learn.model_selection.IncrementalSearch`.

    This estimator does not parallelize the training step. This simply calls
    the underlying estimators's ``fit`` method called and copies over the
    learned attributes to ``self`` afterwards.

    It is helpful for situations where your training dataset is relatively
    small (fits on a single machine) but you need to predict or transform
    a much larger dataset. ``predict``, ``predict_proba`` and ``transform``
    will be done in parallel (potentially distributed if you've connected
    to a Mars cluster).

    Note that many scikit-learn estimators already predict and transform in
    parallel. This meta-estimator may still be useful in those cases when your
    dataset is larger than memory, as the distributed scheduler will ensure the
    data isn't all read into memory at once.

    See Also
    --------
    Incremental
    mars.learn.model_selection.IncrementalSearch

    Examples
    --------
    >>> from sklearn.ensemble import GradientBoostingClassifier
    >>> from sklearn.datasets import make_classification
    >>> import mars.tensor as mt
    >>> from mars.learn.wrappers import ParallelPostFit

    Make a small 1,000 sample 2 training dataset and fit normally.

    >>> X, y = make_classification(n_samples=1000, random_state=0)
    >>> clf = ParallelPostFit(estimator=GradientBoostingClassifier(),
    ...                       scoring='accuracy')
    >>> clf.fit(X, y)
    ParallelPostFit(estimator=GradientBoostingClassifier(...))

    >>> clf.classes_
    array([0, 1])

    Transform and predict return Mars outputs for Mars inputs.

    >>> X_big, y_big = make_classification(n_samples=100000,
                                           random_state=0)
    >>> X_big, y_big = mt.tensor(X_big), mt.tensor(y_big)
    >>> clf.predict(X_big)
    array([1, 0, 0, ..., 1, 0, 0])

    Which can be computed in parallel.

    >>> clf.predict_proba(X_big)
    array([[0.01780031, 0.98219969],
           [0.62199242, 0.37800758],
           [0.89059934, 0.10940066],
           ...,
           [0.03249968, 0.96750032],
           [0.951434  , 0.048566  ],
           [0.99527114, 0.00472886]])
    """

    def __init__(
        self,
        estimator: SklearnBaseEstimator = None,
        scoring: Union[str, Callable] = None,
    ):
        self.estimator = estimator
        self.scoring = scoring

    def _make_fit(self, method):
        def _fit(X, y=None, **kwargs):
            result = (
                mr.spawn(_wrap, args=(self.estimator, method, X, y), kwargs=kwargs)
                .execute()
                .fetch()
            )

            copy_learned_attributes(result, self)
            copy_learned_attributes(result, self.estimator)
            return self

        return _fit

    def fit(self, X, y=None, **kwargs):
        """
        Fit the underlying estimator.

        Parameters
        ----------
        X, y : array-like
        **kwargs
            Additional fit-kwargs for the underlying estimator.

        Returns
        -------
        self : object
        """
        return self._make_fit("fit")(X, y=y, **kwargs)

    def partial_fit(self, X, y=None, **kwargs):  # pragma: no cover
        return self._make_fit("partial_fit")(X, y=y, **kwargs)

    def _check_method(self, method):
        """
        Check if self.estimator has 'method'.

        Raises
        ------
        AttributeError
        """
        estimator = self.estimator
        if not hasattr(estimator, method):
            msg = "The wrapped estimator '{}' does not have a '{}' method.".format(
                estimator, method
            )
            raise AttributeError(msg)
        return getattr(estimator, method)

    def transform(self, X):
        """
        Transform block or partition-wise for Mars inputs.

        For Mars inputs, a Mars tensor is returned. For other
        inputs (NumPy array, pandas dataframe, scipy sparse matrix), the
        regular return value is returned.

        If the underlying estimator does not have a ``transform`` method, then
        an ``AttributeError`` is raised.

        Parameters
        ----------
        X : array-like

        Returns
        -------
        transformed : array-like
        """
        self._check_method("transform")
        X = check_array(X)
        dtype = self.estimator.transform(np.zeros((1, X.shape[1]), dtype=X.dtype)).dtype
        return X.map_chunk(self.estimator.transform, dtype=dtype)

    def score(self, X, y):
        """
        Returns the score on the given data.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
            Input data, where n_samples is the number of samples and
            n_features is the number of features.

        y : array-like, shape = [n_samples] or [n_samples, n_output], optional
            Target relative to X for classification or regression;
            None for unsupervised learning.

        Returns
        -------
        score : float
            return self.estimator.score(X, y)
        """

        scoring = self.scoring
        X = check_array(X)
        y = check_array(y, ensure_2d=False)

        if not scoring:
            if type(self.estimator).score in (
                RegressorMixin.score,
                SklearnRegressorMixin.score,
            ):  # pragma: no cover
                scoring = "r2"
            elif type(self.estimator).score in (
                ClassifierMixin.score,
                SklearnClassifierMixin.score,
            ):
                scoring = "accuracy"
        else:  # pragma: no cover
            scoring = self.scoring

        if scoring:
            scorer = get_scorer(scoring)
            return scorer(self, X, y).execute()
        else:  # pragma: no cover
            return mr.spawn(self.estimator.score, args=(X, y)).execute().fetch()

    def predict(self, X, execute=True):
        """
        Predict for X.

        For Mars inputs, a Mars tensor is returned. For other
        inputs (NumPy array, pandas dataframe, scipy sparse matrix), the
        regular return value is returned.

        Parameters
        ----------
        X : array-like

        Returns
        -------
        y : array-like
        """

        self._check_method("predict")
        X = check_array(X)

        result = X.map_chunk(self.estimator.predict, dtype="int", shape=X.shape[:1])
        if execute:
            result.execute()
        return result

    def predict_proba(self, X, execute=True):
        """
        Probability estimates.

        For Mars inputs, a Mars tensor is returned. For other
        inputs (NumPy array, pandas dataframe, scipy sparse matrix), the
        regular return value is returned.

        If the underlying estimator does not have a ``predict_proba``
        method, then an ``AttributeError`` is raised.

        Parameters
        ----------
        X : array or dataframe

        Returns
        -------
        y : array-like
        """
        self._check_method("predict_proba")
        X = check_array(X)
        result = X.map_chunk(
            self.estimator.predict_proba,
            dtype="float",
            shape=(X.shape[0], len(self.estimator.classes_)),
        )
        if execute:
            result.execute()
        return result

    def predict_log_proba(self, X, execute=True):
        """
        Log of probability estimates.

        For Mars inputs, a Mars tensor is returned. For other
        inputs (NumPy array, pandas dataframe, scipy sparse matrix), the
        regular return value is returned.

        If the underlying estimator does not have a ``predict_proba``
        method, then an ``AttributeError`` is raised.

        Parameters
        ----------
        X : array or dataframe

        Returns
        -------
        y : array-like
        """

        self._check_method("predict_log_proba")
        result = mt.log(self.predict_proba(X, execute=False))
        if execute:
            result.execute()
        return result



# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import functools
import math
import re

from ... import __version__ as mars_version
from ...utils import parse_readable_size, calc_size_by_str

DEFAULT_IMAGE = "marsproject/mars:v" + mars_version
DEFAULT_WORKER_CACHE_MEM = "40%"


def _remove_nones(cfg):
    return dict((k, v) for k, v in cfg.items() if v is not None)


_kube_api_mapping = {
    "v1": "CoreV1Api",
    "apps/v1": "AppsV1Api",
    "rbac.authorization.k8s.io/v1": "RbacAuthorizationV1Api",
}


@functools.lru_cache(10)
def _get_k8s_api(api_version, k8s_api_client):
    from kubernetes import client as kube_client

    return getattr(kube_client, _kube_api_mapping[api_version])(k8s_api_client)


@functools.lru_cache(10)
def _camel_to_underline(name):
    s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()


class KubeConfig(abc.ABC):
    api_version = "v1"

    def create_namespaced(self, k8s_api_client, namespace):
        api = _get_k8s_api(self.api_version, k8s_api_client)
        config = self.build()
        method_name = f'create_namespaced_{_camel_to_underline(config["kind"])}'
        return getattr(api, method_name)(namespace, config)

    @abc.abstractmethod
    def build(self):
        """Build config dict of the object"""


class RoleConfig(KubeConfig):
    """
    Configuration builder for Kubernetes RBAC roles
    """

    api_version = "rbac.authorization.k8s.io/v1"

    def __init__(self, name, namespace, api_groups, resources, verbs):
        self._name = name
        self._namespace = namespace
        self._api_groups = api_groups.split(",")
        self._resources = resources.split(",")
        self._verbs = verbs.split(",")

    def build(self):
        return {
            "kind": "Role",
            "metadata": {"name": self._name, "namespace": self._namespace},
            "rules": [
                {
                    "apiGroups": self._api_groups,
                    "resources": self._resources,
                    "verbs": self._verbs,
                }
            ],
        }


class RoleBindingConfig(KubeConfig):
    """
    Configuration builder for Kubernetes RBAC role bindings
    """

    api_version = "rbac.authorization.k8s.io/v1"

    def __init__(self, name, namespace, role_name, service_account_name):
        self._name = name
        self._namespace = namespace
        self._role_name = role_name
        self._service_account_name = service_account_name

    def build(self):
        return {
            "kind": "RoleBinding",
            "metadata": {"name": self._name, "namespace": self._namespace},
            "roleRef": {
                "apiGroup": "rbac.authorization.k8s.io",
                "kind": "Role",
                "name": self._role_name,
            },
            "subjects": [
                {
                    "kind": "ServiceAccount",
                    "name": self._service_account_name,
                    "namespace": self._namespace,
                }
            ],
        }


class NamespaceConfig(KubeConfig):
    """
    Configuration builder for Kubernetes namespaces
    """

    def __init__(self, name):
        self._name = name

    def build(self):
        return {
            "kind": "Namespace",
            "metadata": {
                "name": self._name,
                "labels": {
                    "name": self._name,
                },
            },
        }


class ServiceConfig(KubeConfig):
    """
    Configuration builder for Kubernetes services
    """

    def __init__(
        self, name, service_type, selector, port, target_port=None, protocol=None
    ):
        self._name = name
        self._type = service_type
        self._protocol = protocol or "TCP"
        self._selector = selector
        self._port = port
        self._target_port = target_port

    def build(self):
        return {
            "kind": "Service",
            "metadata": {
                "name": self._name,
                "labels": {
                    "mars/service-name": self._name,
                },
            },
            "spec": _remove_nones(
                {
                    "type": self._type,
                    "selector": self._selector,
                    "ports": [
                        _remove_nones(
                            {
                                "protocol": self._protocol,
                                "port": self._port,
                                "targetPort": self._target_port,
                            }
                        ),
                    ],
                }
            ),
        }


class ResourceConfig:
    """
    Configuration builder for Kubernetes computation resources
    """

    def __init__(self, cpu, memory):
        self._cpu = cpu
        self._memory, ratio = (
            parse_readable_size(memory) if memory is not None else (None, False)
        )
        assert not ratio

    @property
    def cpu(self):
        return self._cpu

    @property
    def memory(self):
        return self._memory

    def build(self):
        return _remove_nones(
            {
                "cpu": f"{int(self._cpu * 1000)}m" if self._cpu else None,
                "memory": str(int(self._memory)) if self._memory else None,
            }
        )


class PortConfig:
    """
    Configuration builder for Kubernetes ports definition for containers
    """

    def __init__(self, container_port):
        self._container_port = int(container_port)

    def build(self):
        return {
            "containerPort": self._container_port,
        }


class VolumeConfig(abc.ABC):
    """
    Base configuration builder for Kubernetes volumes
    """

    def __init__(self, name, mount_path):
        self.name = name
        self.mount_path = mount_path

    @abc.abstractmethod
    def build(self):
        """Build volume config"""

    def build_mount(self):
        return {
            "name": self.name,
            "mountPath": self.mount_path,
        }


class HostPathVolumeConfig(VolumeConfig):
    """
    Configuration builder for Kubernetes host volumes
    """

    def __init__(self, name, mount_path, host_path, volume_type=None):
        super().__init__(name, mount_path)
        self._host_path = host_path
        self._volume_type = volume_type or "DirectoryOrCreate"

    def build(self):
        return {
            "name": self.name,
            "hostPath": {"path": self._host_path, "type": self._volume_type},
        }


class EmptyDirVolumeConfig(VolumeConfig):
    """
    Configuration builder for Kubernetes empty-dir volumes
    """

    def __init__(self, name, mount_path, use_memory=True, size_limit=None):
        super().__init__(name, mount_path)
        self._medium = "Memory" if use_memory else None
        self._size_limit = size_limit

    def build(self):
        result = {"name": self.name, "emptyDir": {}}
        if self._medium:
            result["emptyDir"]["medium"] = self._medium
        if self._size_limit:
            result["emptyDir"]["sizeLimit"] = str(int(self._size_limit))
        return result


class ContainerEnvConfig:
    """
    Configuration builder for Kubernetes container environments
    """

    def __init__(self, name, value=None, field_path=None):
        self._name = name
        self._value = value
        self._field_path = field_path

    def build(self):
        result = dict(name=self._name)
        if self._value is not None:
            result["value"] = str(self._value)
        elif self._field_path is not None:  # pragma: no branch
            result["valueFrom"] = {"fieldRef": {"fieldPath": self._field_path}}
        return result


class ProbeConfig:
    """
    Base configuration builder for Kubernetes liveness and readiness probes
    """

    def __init__(
        self,
        initial_delay=5,
        period=1,
        timeout=None,
        success_thresh=None,
        failure_thresh=None,
    ):
        self._initial_delay = initial_delay
        self._period = period
        self._timeout = timeout
        self._success_thresh = success_thresh
        self._failure_thresh = failure_thresh

    def build(self):
        return _remove_nones(
            {
                "initialDelaySeconds": self._initial_delay,
                "periodSeconds": self._period,
                "timeoutSeconds": self._timeout,
                "successThreshold": self._success_thresh,
                "failureThreshold": self._failure_thresh,
            }
        )


class TcpSocketProbeConfig(ProbeConfig):
    """
    Configuration builder for TCP liveness and readiness probes
    """

    def __init__(self, port: int, **kwargs):
        super().__init__(**kwargs)
        self._port = port

    def build(self):
        ret = super().build()
        ret["tcpSocket"] = {"port": self._port}
        return ret


class ReplicationConfig(KubeConfig):
    """
    Base configuration builder for Kubernetes replication controllers
    """

    _default_kind = "Deployment"

    def __init__(
        self,
        name,
        image,
        replicas,
        resource_request=None,
        resource_limit=None,
        liveness_probe=None,
        readiness_probe=None,
        pre_stop_command=None,
        kind=None,
    ):
        self._name = name
        self._kind = kind or self._default_kind
        self._image = image
        self._replicas = replicas
        self._ports = []
        self._volumes = []
        self._envs = dict()
        self._labels = dict()

        self.add_default_envs()

        self._resource_request = resource_request
        self._resource_limit = resource_limit

        self._liveness_probe = liveness_probe
        self._readiness_probe = readiness_probe

        self._pre_stop_command = pre_stop_command

    @property
    def api_version(self):
        return "apps/v1" if self._kind in ("Deployment", "ReplicaSet") else "v1"

    def add_env(self, name, value=None, field_path=None):
        self._envs[name] = ContainerEnvConfig(name, value=value, field_path=field_path)

    def remove_env(self, name):  # pragma: no cover
        self._envs.pop(name, None)

    def add_simple_envs(self, envs):
        for k, v in envs.items() or ():
            self.add_env(k, v)

    def add_labels(self, labels):
        self._labels.update(labels)

    def add_port(self, container_port):
        self._ports.append(PortConfig(container_port))

    def add_default_envs(self):
        pass  # pragma: no cover

    def add_volume(self, vol):
        self._volumes.append(vol)

    @abc.abstractmethod
    def build_container_command(self):
        """Output container command"""

    def build_container(self):
        resources_dict = {
            "requests": self._resource_request.build()
            if self._resource_request
            else None,
            "limits": self._resource_limit.build() if self._resource_limit else None,
        }
        lifecycle_dict = _remove_nones(
            {
                "preStop": {
                    "exec": {"command": self._pre_stop_command},
                }
                if self._pre_stop_command
                else None,
            }
        )
        return _remove_nones(
            {
                "command": self.build_container_command(),
                "env": [env.build() for env in self._envs.values()] or None,
                "image": self._image,
                "name": self._name,
                "resources": dict((k, v) for k, v in resources_dict.items() if v)
                or None,
                "ports": [p.build() for p in self._ports] or None,
                "volumeMounts": [vol.build_mount() for vol in self._volumes] or None,
                "livenessProbe": self._liveness_probe.build()
                if self._liveness_probe
                else None,
                "readinessProbe": self._readiness_probe.build()
                if self._readiness_probe
                else None,
                "lifecycle": lifecycle_dict or None,
            }
        )

    def build_template_spec(self):
        result = {
            "containers": [self.build_container()],
            "volumes": [vol.build() for vol in self._volumes],
        }
        return dict((k, v) for k, v in result.items() if v)

    def build(self):
        return {
            "kind": self._kind,
            "metadata": {
                "name": self._name,
            },
            "spec": {
                "replicas": int(self._replicas),
                "template": {
                    "metadata": {
                        "labels": _remove_nones(self._labels) or None,
                    },
                    "spec": self.build_template_spec(),
                },
            },
        }


class MarsReplicationConfig(ReplicationConfig, abc.ABC):
    """
    Base configuration builder for replication controllers for Mars
    """

    rc_name = None
    default_readiness_port = 15031

    def __init__(
        self,
        replicas,
        cpu=None,
        memory=None,
        limit_resources=False,
        memory_limit_ratio=None,
        image=None,
        modules=None,
        volumes=None,
        service_name=None,
        service_port=None,
        **kwargs,
    ):
        self._cpu = cpu
        self._memory, ratio = (
            parse_readable_size(memory) if memory is not None else (None, False)
        )
        assert not ratio

        if isinstance(modules, str):
            self._modules = modules.split(",")
        else:
            self._modules = modules

        req_res = ResourceConfig(cpu, memory) if cpu or memory else None
        limit_res = (
            ResourceConfig(req_res.cpu, req_res.memory * (memory_limit_ratio or 1))
            if req_res and memory
            else None
        )

        self._service_name = service_name
        self._service_port = service_port

        super().__init__(
            self.rc_name,
            image or DEFAULT_IMAGE,
            replicas,
            resource_request=req_res,
            resource_limit=limit_res if limit_resources else None,
            readiness_probe=self.config_readiness_probe(),
            **kwargs,
        )
        if service_port:
            self.add_port(service_port)

        for vol in volumes or ():
            self.add_volume(vol)

        self.add_labels({"mars/service-type": self.rc_name})

    def add_default_envs(self):
        self.add_env("MARS_K8S_POD_NAME", field_path="metadata.name")
        self.add_env("MARS_K8S_POD_NAMESPACE", field_path="metadata.namespace")
        self.add_env("MARS_K8S_POD_IP", field_path="status.podIP")

        if self._service_name:
            self.add_env("MARS_K8S_SERVICE_NAME", str(self._service_name))
        if self._service_port:
            self.add_env("MARS_K8S_SERVICE_PORT", str(self._service_port))

        self.add_env("MARS_CONTAINER_IP", field_path="status.podIP")

        if self._cpu:
            self.add_env("MKL_NUM_THREADS", str(self._cpu))
            self.add_env("MARS_CPU_TOTAL", str(self._cpu))
            if getattr(self, "stat_type", "cgroup") == "cgroup":
                self.add_env("MARS_USE_CGROUP_STAT", "1")

        if self._memory:
            self.add_env("MARS_MEMORY_TOTAL", str(int(self._memory)))

        if self._modules:
            self.add_env("MARS_LOAD_MODULES", ",".join(self._modules))

    def config_readiness_probe(self):
        raise NotImplementedError

    @staticmethod
    def get_local_app_module(mod_name):
        return __name__.rsplit(".", 1)[0] + "." + mod_name

    def build(self):
        result = super().build()
        if self._kind in ("Deployment", "ReplicaSet"):
            result["spec"]["selector"] = {
                "matchLabels": {"mars/service-type": self.rc_name}
            }
        else:
            result["spec"]["selector"] = {"mars/service-type": self.rc_name}
        return result


class MarsSupervisorsConfig(MarsReplicationConfig):
    """
    Configuration builder for Mars supervisor service
    """

    rc_name = "marssupervisor"

    def __init__(self, *args, **kwargs):
        self._web_port = kwargs.pop("web_port", None)
        self._readiness_port = kwargs.pop("readiness_port", self.default_readiness_port)
        super().__init__(*args, **kwargs)
        if self._web_port:
            self.add_port(self._web_port)

    def config_readiness_probe(self):
        return TcpSocketProbeConfig(self._readiness_port, timeout=60, failure_thresh=10)

    def build_container_command(self):
        cmd = [
            "/srv/entrypoint.sh",
            self.get_local_app_module("supervisor"),
        ]
        if self._service_port:
            cmd += ["-p", str(self._service_port)]
        if self._web_port:
            cmd += ["-w", str(self._web_port)]
        if self._cpu:
            cmd += ["--n-process", str(int(math.ceil(self._cpu)))]
        return cmd


class MarsWorkersConfig(MarsReplicationConfig):
    """
    Configuration builder for Mars worker service
    """

    rc_name = "marsworker"

    def __init__(self, *args, **kwargs):
        spill_volumes = kwargs.pop("spill_volumes", None) or ()
        mount_shm = kwargs.pop("mount_shm", True)
        self._limit_resources = kwargs["limit_resources"] = kwargs.get(
            "limit_resources", True
        )
        worker_cache_mem = (
            kwargs.pop("worker_cache_mem", None) or DEFAULT_WORKER_CACHE_MEM
        )
        min_cache_mem = kwargs.pop("min_cache_mem", None)
        self._readiness_port = kwargs.pop("readiness_port", self.default_readiness_port)
        supervisor_web_port = kwargs.pop("supervisor_web_port", None)

        super().__init__(*args, **kwargs)

        self._spill_volumes = []
        for idx, vol in enumerate(spill_volumes):
            if isinstance(vol, str):
                path = f"/mnt/hostpath{idx}"
                self.add_volume(HostPathVolumeConfig(f"host-path-vol-{idx}", path, vol))
                self._spill_volumes.append(path)
            else:
                self.add_volume(vol)
                self._spill_volumes.append(vol.mount_path)
        if self._spill_volumes:
            self.add_env("MARS_SPILL_DIRS", ":".join(self._spill_volumes))

        if self._memory:
            size_limit = calc_size_by_str(worker_cache_mem, self._memory)
            self.add_env("MARS_CACHE_MEM_SIZE", worker_cache_mem)
        else:
            size_limit = None

        if mount_shm:
            self.add_volume(
                EmptyDirVolumeConfig("mars-shared", "/dev/shm", size_limit=size_limit)
            )

        if min_cache_mem:
            self.add_env("MARS_MIN_CACHE_MEM_SIZE", min_cache_mem)
        if supervisor_web_port:
            self.add_env("MARS_K8S_SUPERVISOR_WEB_PORT", supervisor_web_port)

    def config_readiness_probe(self):
        return TcpSocketProbeConfig(self._readiness_port, timeout=60, failure_thresh=10)

    def build_container_command(self):
        cmd = [
            "/srv/entrypoint.sh",
            self.get_local_app_module("worker"),
        ]
        if self._service_port:
            cmd += ["-p", str(self._service_port)]
        return cmd



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

import numpy as np

from ....core import recursive_tile
from ....serialization.serializables import Int64Field
from ....tensor.operands import TensorOperand, TensorOperandMixin
from ....tensor import tensor as astensor
from ....utils import has_unknown_shape
from ...utils import check_array


class PairwiseDistances(TensorOperand, TensorOperandMixin):
    _op_module_ = "learn"

    chunk_store_limit = Int64Field("chunk_store_limit")

    @staticmethod
    def _return_float_dtype(X, Y):
        """
        1. If dtype of X and Y is float32, then dtype float32 is returned.
        2. Else dtype float is returned.
        """

        X = astensor(X)

        if Y is None:
            Y_dtype = X.dtype
        else:
            Y = astensor(Y)
            Y_dtype = Y.dtype

        if X.dtype == Y_dtype == np.float32:
            dtype = np.float32
        else:
            dtype = float

        return X, Y, dtype

    @staticmethod
    def check_pairwise_arrays(X, Y, precomputed=False, dtype=None):
        X, Y, dtype_float = PairwiseDistances._return_float_dtype(X, Y)

        estimator = "check_pairwise_arrays"
        if dtype is None:
            dtype = dtype_float

        if Y is X or Y is None:
            X = Y = check_array(X, accept_sparse=True, dtype=dtype, estimator=estimator)
        else:
            X = check_array(X, accept_sparse=True, dtype=dtype, estimator=estimator)
            Y = check_array(Y, accept_sparse=True, dtype=dtype, estimator=estimator)

        if precomputed:
            if X.shape[1] != Y.shape[0]:
                raise ValueError(
                    "Precomputed metric requires shape "
                    f"(n_queries, n_indexed). Got ({X.shape[0]}, {X.shape[1]}) "
                    f"for {Y.shape[0]} indexed."
                )
        elif X.shape[1] != Y.shape[1]:
            raise ValueError(
                "Incompatible dimension for X and Y matrices: "
                f"X.shape[1] == {X.shape[1]} while Y.shape[1] == {Y.shape[1]}"
            )

        return X, Y

    @classmethod
    def _tile_one_chunk(cls, op):
        out = op.outputs[0]
        chunk_op = op.copy().reset_key()
        chunk = chunk_op.new_chunk(
            [op.x.chunks[0], op.y.chunks[0]],
            shape=out.shape,
            order=out.order,
            index=(0, 0),
        )
        new_op = op.copy()
        return new_op.new_tensors(
            op.inputs,
            shape=out.shape,
            order=out.order,
            chunks=[chunk],
            nsplits=tuple((s,) for s in out.shape),
        )

    @classmethod
    def _tile_chunks(cls, op, x, y):
        out = op.outputs[0]
        out_chunks = []
        for idx in itertools.product(range(x.chunk_shape[0]), range(y.chunk_shape[0])):
            xi, yi = idx

            chunk_op = op.copy().reset_key()
            chunk_inputs = [x.cix[xi, 0], y.cix[yi, 0]]
            out_chunk = chunk_op.new_chunk(
                chunk_inputs,
                shape=(
                    chunk_inputs[0].shape[0],
                    chunk_inputs[1].shape[0],
                ),
                order=out.order,
                index=idx,
            )
            out_chunks.append(out_chunk)

        new_op = op.copy()
        return new_op.new_tensors(
            op.inputs,
            shape=out.shape,
            order=out.order,
            chunks=out_chunks,
            nsplits=(x.nsplits[0], y.nsplits[0]),
        )

    @classmethod
    def _rechunk_cols_into_one(cls, x, y):
        y_is_x = y is x
        if x.chunk_shape[1] != 1 or y.chunk_shape[1] != 1:
            if has_unknown_shape([x, y]):
                yield

            x = yield from recursive_tile(x.rechunk({1: x.shape[1]}))
            if y_is_x:
                y = x
            else:
                y = yield from recursive_tile(y.rechunk({1: y.shape[1]}))

        return x, y

    @classmethod
    def _adjust_chunk_sizes(cls, op, X, Y, out):
        max_x_chunk_size = max(X.nsplits[0])
        max_y_chunk_size = max(Y.nsplits[0])
        itemsize = out.dtype.itemsize
        max_chunk_bytes = max_x_chunk_size * max_y_chunk_size * itemsize
        chunk_store_limit = op.chunk_store_limit * 2  # scale 2 times
        if max_chunk_bytes > chunk_store_limit:
            adjust_succeeded = False
            # chunk is too huge, try to rechunk X and Y
            if X.shape[0] > Y.shape[0]:
                # y is smaller, rechunk y is more efficient
                expected_y_chunk_size = max(
                    int(chunk_store_limit / itemsize / max_x_chunk_size), 1
                )
                if (
                    max_x_chunk_size * expected_y_chunk_size * itemsize
                    <= chunk_store_limit
                ):
                    adjust_succeeded = True
                    Y = yield from recursive_tile(Y.rechunk({0: expected_y_chunk_size}))
            else:
                # x is smaller, rechunk x is more efficient
                expected_x_chunk_size = max(
                    int(chunk_store_limit / itemsize / max_y_chunk_size), 1
                )
                if (
                    max_y_chunk_size * expected_x_chunk_size * itemsize
                    <= chunk_store_limit
                ):
                    adjust_succeeded = True
                    X = yield from recursive_tile(X.rechunk({0: expected_x_chunk_size}))

            if not adjust_succeeded:
                expected_chunk_size = max(int(np.sqrt(chunk_store_limit / itemsize)), 1)
                X = yield from recursive_tile(X.rechunk({0: expected_chunk_size}))
                Y = yield from recursive_tile(Y.rechunk({0: expected_chunk_size}))

        return X, Y



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union, List, Dict, Any

from .core import Serializer, buffered

try:
    import pyarrow as pa

    pa_types = Union[pa.Table, pa.RecordBatch]
except ImportError:  # pragma: no cover
    pa = None
    pa_types = Any


class ArrowBatchSerializer(Serializer):
    @buffered
    def serial(self, obj: pa_types, context: Dict):
        sink = pa.BufferOutputStream()
        writer = pa.RecordBatchStreamWriter(sink, obj.schema)
        if isinstance(obj, pa.Table):
            batch_type = "T"
            writer.write_table(obj)
        else:
            batch_type = "B"
            writer.write_batch(obj)
        writer.close()

        buf = sink.getvalue()
        buffers = [buf]
        return (batch_type,), buffers, True

    def deserial(self, serialized: Dict, context: Dict, subs: List):
        reader = pa.RecordBatchStreamReader(pa.BufferReader(subs[0]))
        if serialized[0] == "T":
            return reader.read_all()
        else:
            return reader.read_next_batch()


if pa is not None:  # pragma: no branch
    ArrowBatchSerializer.register(pa.Table)
    ArrowBatchSerializer.register(pa.RecordBatch)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

import numpy as np
import pandas as pd
from pandas.api.types import is_list_like

from ... import opcodes as OperandDef
from ...core import ENTITY_TYPE
from ...serialization.serializables import KeyField, AnyField
from ...tensor.core import TENSOR_TYPE
from ..core import DATAFRAME_TYPE, SERIES_TYPE, INDEX_TYPE
from ..operands import DataFrameOperand, DataFrameOperandMixin


class DataFrameIsin(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.ISIN

    input = KeyField("input")
    values = AnyField("values")

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        inputs_iter = iter(self._inputs)
        self.input = next(inputs_iter)
        if len(self._inputs) > 1:
            if isinstance(self.values, dict):
                new_values = dict()
                for k, v in self.values.items():
                    if isinstance(v, ENTITY_TYPE):
                        new_values[k] = next(inputs_iter)
                    else:
                        new_values[k] = v
                self.values = new_values
            else:
                self.values = self._inputs[1]

    def __call__(self, elements):
        inputs = [elements]
        if isinstance(self.values, ENTITY_TYPE):
            inputs.append(self.values)
        elif isinstance(self.values, dict):
            for v in self.values.values():
                if isinstance(v, ENTITY_TYPE):
                    inputs.append(v)

        if elements.ndim == 1:
            return self.new_series(
                inputs,
                shape=elements.shape,
                dtype=np.dtype("bool"),
                index_value=elements.index_value,
                name=elements.name,
            )
        else:
            dtypes = pd.Series(
                [np.dtype(bool) for _ in elements.dtypes], index=elements.dtypes.index
            )
            return self.new_dataframe(
                inputs,
                shape=elements.shape,
                index_value=elements.index_value,
                columns_value=elements.columns_value,
                dtypes=dtypes,
            )

    @classmethod
    def _tile_entity_values(cls, op):
        from ..utils import auto_merge_chunks
        from ..arithmetic.bitwise_or import tree_dataframe_or
        from ...core.context import get_context

        in_elements = op.input
        out_elements = op.outputs[0]
        # values contains mars objects
        chunks_list = []
        in_chunks = in_elements.chunks
        if any(len(t.chunks) > 4 for t in op.inputs):
            # yield and merge value chunks to reduce graph nodes
            yield list(
                itertools.chain.from_iterable(
                    t.chunks for t in op.inputs if isinstance(t, ENTITY_TYPE)
                )
            )
            in_elements = auto_merge_chunks(get_context(), op.input)
            in_chunks = in_elements.chunks
            for value in op.inputs[1:]:
                if isinstance(value, DATAFRAME_TYPE + SERIES_TYPE):
                    merged = auto_merge_chunks(get_context(), value)
                    chunks_list.append(merged.chunks)
                elif isinstance(value, ENTITY_TYPE):
                    chunks_list.append(value.chunks)
        else:
            for value in op.inputs[1:]:
                if isinstance(value, ENTITY_TYPE):
                    chunks_list.append(value.chunks)

        out_chunks = []
        for in_chunk in in_chunks:
            isin_chunks = []
            for value_chunks in itertools.product(*chunks_list):
                input_chunks = [in_chunk] + list(value_chunks)
                isin_chunks.append(cls._new_chunk(op, in_chunk, input_chunks))
            out_chunk = tree_dataframe_or(*isin_chunks, index=in_chunk.index)
            out_chunks.append(out_chunk)

        new_op = op.copy()
        params = out_elements.params
        params["nsplits"] = in_elements.nsplits
        params["chunks"] = out_chunks
        return new_op.new_tileables(op.inputs, kws=[params])

    @classmethod
    def tile(cls, op):
        in_elements = op.input
        out_elements = op.outputs[0]

        if len(op.inputs) > 1:
            return (yield from cls._tile_entity_values(op))

        out_chunks = []
        for chunk in in_elements.chunks:
            out_chunk = cls._new_chunk(op, chunk, [chunk])
            out_chunks.append(out_chunk)

        new_op = op.copy()
        params = out_elements.params
        params["nsplits"] = in_elements.nsplits
        params["chunks"] = out_chunks
        return new_op.new_tileables(op.inputs, kws=[params])

    @classmethod
    def _new_chunk(cls, op, chunk, input_chunks):
        out_elements = op.outputs[0]
        chunk_op = op.copy().reset_key()
        if out_elements.ndim == 1:
            out_chunk = chunk_op.new_chunk(
                input_chunks,
                shape=chunk.shape,
                dtype=out_elements.dtype,
                index_value=chunk.index_value,
                name=out_elements.name,
                index=chunk.index,
            )
        else:
            chunk_dtypes = pd.Series(
                [np.dtype(bool) for _ in chunk.dtypes], index=chunk.dtypes.index
            )
            out_chunk = chunk_op.new_chunk(
                input_chunks,
                shape=chunk.shape,
                index_value=chunk.index_value,
                columns_value=chunk.columns_value,
                dtypes=chunk_dtypes,
                index=chunk.index,
            )
        return out_chunk

    @classmethod
    def execute(cls, ctx, op):
        inputs_iter = iter(op.inputs)
        elements = ctx[next(inputs_iter).key]

        if isinstance(op.values, dict):
            values = dict()
            for k, v in op.values.items():
                if isinstance(v, ENTITY_TYPE):
                    values[k] = ctx[next(inputs_iter).key]
                else:
                    values[k] = v
        else:
            if isinstance(op.values, ENTITY_TYPE):
                values = ctx[next(inputs_iter).key]
            else:
                values = op.values

        try:
            ctx[op.outputs[0].key] = elements.isin(values)
        except ValueError:
            # buffer read-only
            ctx[op.outputs[0].key] = elements.copy().isin(values.copy())


def series_isin(elements, values):
    """
    Whether elements in Series are contained in `values`.

    Return a boolean Series showing whether each element in the Series
    matches an element in the passed sequence of `values` exactly.

    Parameters
    ----------
    values : set or list-like
        The sequence of values to test. Passing in a single string will
        raise a ``TypeError``. Instead, turn a single string into a
        list of one element.

    Returns
    -------
    Series
        Series of booleans indicating if each element is in values.

    Raises
    ------
    TypeError
      * If `values` is a string

    See Also
    --------
    DataFrame.isin : Equivalent method on DataFrame.

    Examples
    --------
    >>> import mars.dataframe as md
    >>> s = md.Series(['lama', 'cow', 'lama', 'beetle', 'lama',
    ...                'hippo'], name='animal')
    >>> s.isin(['cow', 'lama']).execute()
    0     True
    1     True
    2     True
    3    False
    4     True
    5    False
    Name: animal, dtype: bool

    Passing a single string as ``s.isin('lama')`` will raise an error. Use
    a list of one element instead:

    >>> s.isin(['lama']).execute()
    0     True
    1    False
    2     True
    3    False
    4     True
    5    False
    Name: animal, dtype: bool
    """
    if is_list_like(values):
        values = list(values)
    elif not isinstance(values, (SERIES_TYPE, TENSOR_TYPE, INDEX_TYPE)):
        raise TypeError(
            "only list-like objects are allowed to be passed to isin(), "
            f"you passed a [{type(values)}]"
        )
    op = DataFrameIsin(values=values)
    return op(elements)


def df_isin(df, values):
    """
    Whether each element in the DataFrame is contained in values.

    Parameters
    ----------
    values : iterable, Series, DataFrame or dict
        The result will only be true at a location if all the
        labels match. If `values` is a Series, that's the index. If
        `values` is a dict, the keys must be the column names,
        which must match. If `values` is a DataFrame,
        then both the index and column labels must match.

    Returns
    -------
    DataFrame
        DataFrame of booleans showing whether each element in the DataFrame
        is contained in values.

    See Also
    --------
    DataFrame.eq: Equality test for DataFrame.
    Series.isin: Equivalent method on Series.
    Series.str.contains: Test if pattern or regex is contained within a
        string of a Series or Index.

    Examples
    --------
    >>> import mars.dataframe as md
    >>> df = md.DataFrame({'num_legs': [2, 4], 'num_wings': [2, 0]},
    ...                   index=['falcon', 'dog'])
    >>> df.execute()
            num_legs  num_wings
    falcon         2          2
    dog            4          0

    When ``values`` is a list check whether every value in the DataFrame
    is present in the list (which animals have 0 or 2 legs or wings)

    >>> df.isin([0, 2]).execute()
            num_legs  num_wings
    falcon      True       True
    dog        False       True

    When ``values`` is a dict, we can pass values to check for each
    column separately:

    >>> df.isin({'num_wings': [0, 3]}).execute()
            num_legs  num_wings
    falcon     False      False
    dog        False       True

    When ``values`` is a Series or DataFrame the index and column must
    match. Note that 'falcon' does not match based on the number of legs
    in df2.

    >>> other = md.DataFrame({'num_legs': [8, 2], 'num_wings': [0, 2]},
    ...                      index=['spider', 'falcon'])
    >>> df.isin(other).execute()
            num_legs  num_wings
    falcon      True       True
    dog        False      False
    """
    if is_list_like(values) and not isinstance(values, dict):
        values = list(values)
    elif not isinstance(
        values, (SERIES_TYPE, DATAFRAME_TYPE, TENSOR_TYPE, INDEX_TYPE, dict)
    ):
        raise TypeError(
            "only list-like objects or dict are allowed to be passed to isin(), "
            f"you passed a [{type(values)}]"
        )
    op = DataFrameIsin(values=values)
    return op(df)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import importlib
import logging
import time
from collections import defaultdict
from dataclasses import dataclass
from typing import Any, Dict, List, Type

from .... import oscar as mo
from ....core import TileableGraph, TileableType, enter_mode, TileContext
from ....core.operand import Fetch
from ...subtask import SubtaskResult, SubtaskGraph
from ..config import task_options
from ..core import Task, new_task_id, TaskStatus
from ..errors import TaskNotExist
from .preprocessor import TaskPreprocessor
from .processor import TaskProcessor
from .task import TaskProcessorActor

logger = logging.getLogger(__name__)


class TaskConfigurationActor(mo.Actor):
    def __init__(
        self,
        task_conf: Dict[str, Any],
        execution_config: Dict[str, Any],
        task_processor_cls: Type[TaskProcessor] = None,
        task_preprocessor_cls: Type[TaskPreprocessor] = None,
    ):
        for name, value in task_conf.items():
            setattr(task_options, name, value)
        self._execution_config = execution_config
        self._task_processor_cls = task_processor_cls
        self._task_preprocessor_cls = task_preprocessor_cls

    def get_config(self):
        return {
            "task_options": task_options,
            "execution_config": self._execution_config,
            "task_processor_cls": self._task_processor_cls,
            "task_preprocessor_cls": self._task_preprocessor_cls,
        }


@dataclass
class ResultTileableInfo:
    tileable: TileableType
    processor_ref: mo.ActorRefType[TaskProcessorActor]


class TaskManagerActor(mo.Actor):
    _task_name_to_parent_task_id: Dict[str, str]
    _task_name_to_task_ids: Dict[str, List[str]]

    _task_id_to_processor_ref: Dict[str, mo.ActorRefType[TaskProcessorActor]]
    _tileable_key_to_info: Dict[str, List[ResultTileableInfo]]

    def __init__(self, session_id: str):
        self._session_id = session_id

        self._config = None
        self._execution_config = None
        self._task_processor_cls = None
        self._task_preprocessor_cls = None
        self._last_idle_time = None

        self._task_name_to_parent_task_id = dict()
        self._task_name_to_task_ids = defaultdict(list)

        self._task_id_to_processor_ref = dict()
        self._tileable_key_to_info = defaultdict(list)

    async def __post_create__(self):
        # get config
        configuration_ref = await mo.actor_ref(
            TaskConfigurationActor.default_uid(), address=self.address
        )
        task_conf = await configuration_ref.get_config()
        (
            self._config,
            self._execution_config,
            self._task_processor_cls,
            self._task_preprocessor_cls,
        ) = (
            task_conf["task_options"],
            task_conf["execution_config"],
            task_conf["task_processor_cls"],
            task_conf["task_preprocessor_cls"],
        )
        self._task_preprocessor_cls = self._get_task_preprocessor_cls()

    async def __pre_destroy__(self):
        for processor_ref in self._task_id_to_processor_ref.values():
            await processor_ref.destroy()

    @staticmethod
    def gen_uid(session_id):
        return f"{session_id}_task_manager"

    @enter_mode(kernel=True)
    async def submit_tileable_graph(
        self,
        graph: TileableGraph,
        task_name: str = None,
        fuse_enabled: bool = None,
        extra_config: dict = None,
    ) -> str:
        self._last_idle_time = None
        if task_name is None:
            # new task without task name
            task_id = task_name = new_task_id()
            parent_task_id = new_task_id()
        elif task_name in self._task_name_to_parent_task_id:
            # task with the same name submitted before
            parent_task_id = self._task_name_to_parent_task_id[task_name]
            task_id = new_task_id()
        else:
            # new task with task_name
            task_id = new_task_id()
            parent_task_id = new_task_id()

        uid = TaskProcessorActor.gen_uid(self._session_id, parent_task_id)
        if task_name not in self._task_name_to_parent_task_id:
            # gen main task which mean each submission from user
            processor_ref = await mo.create_actor(
                TaskProcessorActor,
                self._session_id,
                parent_task_id,
                task_name=task_name,
                task_processor_cls=self._task_processor_cls,
                address=self.address,
                uid=uid,
            )
            self._task_name_to_parent_task_id[task_name] = parent_task_id
        else:
            processor_ref = await mo.actor_ref(mo.ActorRef(self.address, uid))
        self._task_name_to_task_ids[task_name].append(task_id)
        self._task_id_to_processor_ref[task_id] = processor_ref

        if fuse_enabled is None:
            fuse_enabled = self._config.fuse_enabled
        # gen task
        task = Task(
            task_id,
            self._session_id,
            graph,
            task_name,
            parent_task_id=parent_task_id,
            fuse_enabled=fuse_enabled,
            extra_config=extra_config,
        )
        # gen task processor
        tiled_context = await self._gen_tiled_context(graph)
        await processor_ref.add_task(
            task,
            tiled_context,
            self._config,
            self._execution_config,
            self._task_preprocessor_cls,
        )

        for tileable in graph.result_tileables:
            info = ResultTileableInfo(tileable=tileable, processor_ref=processor_ref)
            self._tileable_key_to_info[tileable.key].append(info)

        return task_id

    async def get_subtask_graphs(self, task_id: str) -> List[SubtaskGraph]:
        try:
            processor_ref = self._task_id_to_processor_ref[task_id]
        except KeyError:  # pragma: no cover
            raise TaskNotExist(f"Task {task_id} does not exist")

        return processor_ref.get_subtask_graphs(task_id)

    async def get_tileable_graph_dict_by_task_id(self, task_id: str):
        try:
            processor_ref = self._task_id_to_processor_ref[task_id]
        except KeyError:
            raise TaskNotExist(f"Task {task_id} does not exist")

        res = await processor_ref.get_tileable_graph_as_dict()
        return res

    async def get_tileable_details(self, task_id):
        try:
            processor_ref = self._task_id_to_processor_ref[task_id]
        except KeyError:
            raise TaskNotExist(f"Task {task_id} does not exist")

        return await processor_ref.get_tileable_details()

    async def get_tileable_subtasks(self, task_id, tileable_id, with_input_output):
        try:
            processor_ref = self._task_id_to_processor_ref[task_id]
        except KeyError:
            raise TaskNotExist(f"Task {task_id} does not exist")

        return await processor_ref.get_tileable_subtasks(tileable_id, with_input_output)

    async def _gen_tiled_context(self, graph: TileableGraph) -> TileContext:
        # process graph, add fetch node to tiled context
        tiled_context = TileContext()
        for tileable in graph:
            if isinstance(tileable.op, Fetch) and tileable.is_coarse():
                info = self._tileable_key_to_info[tileable.key][-1]
                tiled_context[tileable] = await info.processor_ref.get_result_tileable(
                    tileable.key
                )
        return tiled_context

    def _get_task_preprocessor_cls(self):
        if self._task_preprocessor_cls is not None:
            assert isinstance(self._task_preprocessor_cls, str)
            module, name = self._task_preprocessor_cls.rsplit(".", 1)
            return getattr(importlib.import_module(module), name)
        else:
            return TaskPreprocessor

    async def wait_task(self, task_id: str, timeout: int = None):
        try:
            processor_ref = self._task_id_to_processor_ref[task_id]
        except KeyError:  # pragma: no cover
            raise TaskNotExist(f"Task {task_id} does not exist")

        return processor_ref.wait(timeout)

    async def cancel_task(self, task_id: str):
        try:
            processor_ref = self._task_id_to_processor_ref[task_id]
        except KeyError:  # pragma: no cover
            raise TaskNotExist(f"Task {task_id} does not exist")

        yield processor_ref.cancel()

    async def get_task_results(self, progress: bool = False):
        if not self._task_id_to_processor_ref:
            raise mo.Return([])

        results = yield asyncio.gather(
            *[ref.result() for ref in self._task_id_to_processor_ref.values()]
        )

        if progress:
            task_to_result = {res.task_id: res for res in results}

            progress_task_ids = []
            for res in results:
                if res.status != TaskStatus.terminated:
                    progress_task_ids.append(res.task_id)
                else:
                    res.progress = 1.0

            progresses = yield asyncio.gather(
                *[
                    self._task_id_to_processor_ref[task_id].progress()
                    for task_id in progress_task_ids
                ]
            )
            for task_id, progress in zip(progress_task_ids, progresses):
                task_to_result[task_id].progress = progress

        raise mo.Return(results)

    async def get_task_result(self, task_id: str):
        try:
            processor_ref = self._task_id_to_processor_ref[task_id]
        except KeyError:  # pragma: no cover
            raise TaskNotExist(f"Task {task_id} does not exist")

        return await processor_ref.result()

    async def get_task_result_tileables(self, task_id: str):
        try:
            processor_ref = self._task_id_to_processor_ref[task_id]
        except KeyError:  # pragma: no cover
            raise TaskNotExist(f"Task {task_id} does not exist")

        return await processor_ref.get_result_tileables()

    async def set_subtask_result(self, subtask_result: SubtaskResult):
        task_id = subtask_result.task_id
        try:
            processor_ref = self._task_id_to_processor_ref[task_id]
        except KeyError:  # pragma: no cover
            # raise TaskNotExist(f'Task {task_id} does not exist')
            logger.warning(
                "Current task is finished, got stale result %s  for subtask %s "
                "which may be speculative execution from previous tasks, just ignore it.",
                subtask_result.subtask_id,
                subtask_result,
            )
            return

        yield processor_ref.set_subtask_result(subtask_result)

    @mo.extensible
    async def get_task_progress(self, task_id: str) -> float:
        try:
            processor_ref = self._task_id_to_processor_ref[task_id]
        except KeyError:  # pragma: no cover
            raise TaskNotExist(f"Task {task_id} does not exist")

        return await processor_ref.progress()

    async def get_last_idle_time(self):
        if self._last_idle_time is None:
            for processor_ref in self._task_id_to_processor_ref.values():
                if not await processor_ref.is_done():
                    break
            else:
                self._last_idle_time = time.time()
        return self._last_idle_time
